from threemystic_cloud_data_client.cloud_providers.azure.client.actions.base_class.base import cloud_data_client_azure_client_action_base as base
import asyncio
from azure.mgmt.compute import ComputeManagementClient



class cloud_data_client_azure_client_action(base):
  def __init__(self, *args, **kwargs):
    super().__init__(
      data_action="vmimage", 
      logger_name= "cloud_data_client_azure_client_action_vmimage",
      *args, **kwargs)
      
  async def _process_account_data(self, account, loop, *args, **kwargs):
    client = ComputeManagementClient(credential= self.get_cloud_client().get_tenant_credential(tenant= self.get_cloud_client().get_tenant_id(tenant= account, is_account= True)), subscription_id= self.get_cloud_client().get_account_id(account= account))
    image_galleries = self.get_cloud_client().sdk_request(
           tenant= self.get_cloud_client().get_tenant_id(tenant= account, is_account= True), 
           lambda_sdk_command=lambda: client.galleries.list()
          )
    
    image_gallery_images = { }
    
    for gallery in image_galleries:
      for image in self.get_cloud_client().sdk_request(
          tenant= self.get_cloud_client().get_tenant_id(tenant= account, is_account= True), 
          lambda_sdk_command=lambda: client.gallery_images.list_by_gallery(
            resource_group_name= self.get_cloud_client().get_resource_group_from_resource(resource= gallery),
            gallery_name= gallery.name
          )
        ):
        image_gallery_images[self.get_cloud_client().get_resource_id(resource= gallery)] = self.get_common().helper_type().dictionary().merge_dictionary([
          {},
          {"extra_versions": [
            self.get_cloud_client().serialize_resource(resource= image_version) 
            for image_version in self.get_cloud_client().sdk_request(
            tenant= self.get_cloud_client().get_tenant_id(tenant= account, is_account= True), 
            lambda_sdk_command=lambda: client.gallery_image_versions.list_by_gallery_image(
                resource_group_name= self.get_cloud_client().get_resource_group_from_resource(resource= gallery),
                gallery_name= gallery.name,
                gallery_image_name= image.name
              )
            )
          ]},
          self.get_cloud_client().serialize_resource(resource= image)
          
        ])


    return {
      "account": account,
      "data": [
          self.get_common().helper_type().dictionary().merge_dictionary([
          {},
          await self.get_base_return_data(
            account= self.get_cloud_client().serialize_resource(resource= account),
            resource_id= self.get_cloud_client().get_resource_id(resource= item),
            resource= item,
            region= self.get_cloud_client().get_resource_location(resource= item),
            resource_groups= [self.get_cloud_client().get_resource_group_from_resource(resource= item)],
          ),
          image_gallery_images[self.get_cloud_client().get_resource_id(resource= item)],
        ]) for item in image_gallery_images
      ]
    }


"""The AWS database Action. This will pull the AWS rds"""
from threemystic_cloud_data_client.cloud_providers.aws.client.actions.base_class.base import cloud_data_client_aws_client_action_base as base
import asyncio
from decimal import Decimal, ROUND_HALF_UP
from random import randint

class cloud_data_client_aws_client_action(base):
  def __init__(self, *args, **kwargs):
    super().__init__(
      data_action="budget",
      logger_name= "cloud_data_client_aws_client_action_budget",
      *args, **kwargs)
  
  async def _process_account_data_region(self, account, region, resource_groups, loop, *args, **kwargs):
    pass

  def __init_costdata_month(self, data_dt, *args, **kwargs):
    return {
      "currency": self.get_common().helper_type().string().set_case(string_value= self.get_cloud_data_client().get_default_currency(), case= "upper"),
      "month": data_dt.month,
      "year": data_dt.year,
      "totals":{
        "total": Decimal(0),
        "fiscal_total": Decimal(0),
        "forcast_total": Decimal(0),
        "fiscal_forcast_total": Decimal(0),
        "resource_group": {
          "total":{},
          "forcast_total":{},
          "origional_currency_total":{},
          "origional_currency_forcast_total":{},
        },
        "resource_type": {
          "total":{},
          "forcast_total":{},
          "origional_currency_total":{},
          "origional_currency_forcast_total":{},
        }
      },
      "days":{}
    }
  
  def __init_costdata_month_day(self, data_dt, currency, *args, **kwargs):
    return {
      "currency": self.get_common().helper_type().string().set_case(string_value= self.get_cloud_data_client().get_default_currency(), case= "upper"),
      "origional_currency": self.get_common().helper_type().string().set_case(string_value= currency, case= "upper"),
      "date": data_dt,
      "total": Decimal(0),
      "forcast_total": Decimal(0),
      "origional_currency_total": Decimal(0),
      "origional_currency_forcast_total": Decimal(0),
      "resource_group": {
        "total":{},
        "forcast_total":{},
        "origional_currency_total":{},
        "origional_currency_forcast_total":{},
      },
      "resource_type": {
        "total":{},
        "forcast_total":{},
        "origional_currency_total":{},
        "origional_currency_forcast_total":{},
      }
    }
  
  def __convert_costmetric_forecast_metric(self, cost_metric, *args, **kwargs):
    if self.get_common().helper_type().string().set_case(string_value= cost_metric, case= "lower") == "netunblendedcost":
      return "NET_UNBLENDED_COST"
    
    if self.get_common().helper_type().string().set_case(string_value= cost_metric, case= "lower") == "blendedcost":
      return "BLENDED_COST"
    
    if self.get_common().helper_type().string().set_case(string_value= cost_metric, case= "lower") == "unblendedcost":
      return "UNBLENDED_COST"
    
    if self.get_common().helper_type().string().set_case(string_value= cost_metric, case= "lower") == "amortizedcost":
      return "AMORTIZED_COST"
    
    if self.get_common().helper_type().string().set_case(string_value= cost_metric, case= "lower") == "netamortizedcost":
      return "NET_AMORTIZED_COST"
    
    if self.get_common().helper_type().string().set_case(string_value= cost_metric, case= "lower") == "usagequantity":
      return "USAGE_QUANTITY"
    
    if self.get_common().helper_type().string().set_case(string_value= cost_metric, case= "lower") == "normalizedusageamount":
      return "NORMALIZED_USAGE_AMOUNT"
    
    return cost_metric

  def __process_get_cost_data_process_forcast_process_day(self, cost_metric, year_data, data_dt, fiscal_start, fiscal_end, raw_data_cost, currency, total_key, *args, **kwargs):
    by_month_key = self.get_common().helper_type().datetime().datetime_as_string(dt_format= "%Y%m", dt= data_dt)
    day_key = self.get_common().helper_type().datetime().datetime_as_string(dt_format= "%Y%m%d", dt= data_dt)

    if year_data.get(by_month_key) is None:
      year_data[by_month_key] = self.__init_costdata_month(data_dt= data_dt)
    
    if year_data[by_month_key]["days"].get(day_key) is None:
      year_data[by_month_key]["days"][day_key] = self.__init_costdata_month_day(data_dt= data_dt, currency= currency)
    
    raw_row_data_cost = raw_data_cost
    row_data_cost = raw_data_cost
    
    if year_data[by_month_key]["days"][day_key]["currency"] != year_data[by_month_key]["days"][day_key]["origional_currency"]:
      row_data_cost = self.get_common().helper_currency().convert(
        ammount= row_data_cost,
        currency_from= currency,
        currency_to= year_data[by_month_key]["days"][day_key]["currency"],
        conversion_date= self.get_common().helper_type().datetime().yesterday(dt=self.get_common().helper_type().datetime().datetime_from_string(
          dt_string= self.get_common().helper_type().datetime().datetime_as_string(
            dt= data_dt,
            dt_format= "%Y%m01"
          ),
          dt_format= "%Y%m%d"
        )).date()
      )

    year_data[by_month_key]["days"][day_key][f'origional_currency_{total_key}'] += Decimal(raw_row_data_cost)
    year_data[by_month_key]["days"][day_key][f'{total_key}'] += Decimal(row_data_cost)
    year_data[by_month_key]["totals"][f'{total_key}'] += Decimal(row_data_cost)
    
    if data_dt >= fiscal_start and data_dt <= fiscal_end:
      year_data[by_month_key]["totals"][f'fiscal_{total_key}'] += Decimal(row_data_cost)
  
  async def __process_get_cost_data_process_forcast(self, year_data, client, account, start_date, end_date, fiscal_start, fiscal_end, cost_metrics, *args, **kwargs):
    
    adjusted_enddated = (start_date
                 + self.get_common().helper_type().datetime().time_delta(months= 4, dt= start_date))
    for cost_metric in cost_metrics:
      
      adjusted_enddated = (start_date + self.get_common().helper_type().datetime().time_delta(months= 4, dt= start_date))
      adjusted_enddated = self.get_common().helper_type().datetime().datetime_from_string(adjusted_enddated.strftime("%Y-%m-01"), dt_format="%Y-%m-%d")
      if end_date < adjusted_enddated:
       adjusted_enddated =  end_date
      
      results_by_time_forcast = self.get_cloud_client().general_boto_call_single(
        boto_call=lambda: client.get_cost_forecast(
          TimePeriod={
            'Start': start_date.strftime("%Y-%m-%d"),
            'End': adjusted_enddated.strftime("%Y-%m-%d"),
          },
          Granularity='DAILY',
          Metric=self.__convert_costmetric_forecast_metric(cost_metric= cost_metric),
          Filter={
            "Dimensions":{
              "Key":"LINKED_ACCOUNT",
              "Values":[self.get_cloud_client().get_account_id(account= account)]
            }
          },
        ),
        boto_params= None,
        boto_nextkey = None,
        boto_key= None
      )
      
      total_key = "forcast_total"
      currency = results_by_time_forcast["Total"]["Unit"]
    
    
      year_data[cost_metric] = {}
      data_dt = None
      for cost_data in results_by_time_forcast["ForecastResultsByTime"]:
        data_dt = self.get_common().helper_type().datetime().datetime_from_string(dt_string= str(cost_data["TimePeriod"]["Start"]), dt_format= "%Y-%m-%d")

        self.__process_get_cost_data_process_forcast_process_day(
          cost_metric= cost_metric,
          year_data= year_data[cost_metric],
          data_dt= data_dt,
          fiscal_start= fiscal_start,
          fiscal_end= fiscal_end,
          raw_data_cost = (cost_data["MeanValue"]),
          currency= currency,
          total_key= total_key
        )

      if data_dt is not None:
        by_month_key = self.get_common().helper_type().datetime().datetime_as_string(dt_format= "%Y%m", dt= data_dt)
        day_key = self.get_common().helper_type().datetime().datetime_as_string(dt_format= "%Y%m%d", dt= data_dt)
        cost_data = year_data[cost_metric][by_month_key]["days"][day_key]

        while data_dt < end_date:
          data_dt += (self.get_common().helper_type().datetime().time_delta(days= 1, dt= data_dt))         

          self.__process_get_cost_data_process_forcast_process_day(
            cost_metric= cost_metric,
            year_data= year_data[cost_metric],
            data_dt= data_dt,
            fiscal_start= fiscal_start,
            fiscal_end= fiscal_end,
            raw_data_cost = (cost_data[total_key] * ((randint(90,120)/Decimal(100)))),
            currency= (cost_data["currency"]),
            total_key= total_key
          )


  def get_total_cost_data(self, cost_data, cost_metric, *args, **kwargs):
    if cost_data["Total"].get(cost_metric) is not None:
      if cost_data["Total"][cost_metric].get("Amount") is not None:
        return Decimal(cost_data["Total"][cost_metric]["Amount"])
    
    return Decimal(0)
      
  def get_currency_cost_data(self, cost_data, cost_metric, *args, **kwargs):
    if cost_data["Total"].get(cost_metric) is not None:
      if cost_data["Total"][cost_metric].get("Unit") is not None:
        return cost_data["Total"][cost_metric]["Unit"]
    
    if len(cost_data["Groups"]) > 0:
      if cost_data["Groups"][0]["Metrics"].get(cost_metric) is not None:
        if cost_data["Groups"][0]["Metrics"][cost_metric].get("Unit") is not None:
          return cost_data["Groups"][0]["Metrics"][cost_metric]["Unit"]
        
    return self.get_cloud_data_client().get_default_currency()
      
  async def __process_get_cost_data_process_year_data(self, year_data, client, account, start_date, end_date, fiscal_start, fiscal_end, cost_metrics, *args, **kwargs):

    results_by_time = self.get_cloud_client().general_boto_call_array(
      boto_call=lambda item: client.get_cost_and_usage(
        TimePeriod={
          'Start': start_date.strftime("%Y-%m-%d"),
          'End': end_date.strftime("%Y-%m-%d"),
        },
        Granularity='DAILY',
        Filter={
          "Dimensions":{
            "Key":"LINKED_ACCOUNT",
            "Values":[self.get_cloud_client().get_account_id(account= account)]
          }
        },
        GroupBy= [
          {
            "Type": "DIMENSION",
            "Key": "SERVICE" 
          }
        ],
        **item
      ),
      boto_params= {
        "Metrics": cost_metrics
      },
      boto_nextkey = "NextPageToken",
      boto_nextkey_param = "NextPageToken",
      boto_key="ResultsByTime"
    )
    
    total_key = "total"
    for cost_metric in cost_metrics:
      year_data[cost_metric] = {}
      for cost_data in results_by_time:
        data_dt = self.get_common().helper_type().datetime().datetime_from_string(dt_string= str(cost_data["TimePeriod"]["Start"]), dt_format= "%Y-%m-%d")
        by_month_key = self.get_common().helper_type().datetime().datetime_as_string(dt_format= "%Y%m", dt= data_dt)

        day_key = self.get_common().helper_type().datetime().datetime_as_string(dt_format= "%Y%m%d", dt= data_dt)
        if year_data[cost_metric].get(by_month_key) is None:
          year_data[cost_metric][by_month_key] = self.__init_costdata_month(data_dt= data_dt)
        
        if year_data[cost_metric][by_month_key]["days"].get(day_key) is None:
          year_data[cost_metric][by_month_key]["days"][day_key] = self.__init_costdata_month_day(data_dt= data_dt, currency= self.get_currency_cost_data(cost_data= cost_data, cost_metric= cost_metric))

        raw_row_data_cost = self.get_total_cost_data(cost_data= cost_data, cost_metric= cost_metric)
        total_attribute_empty = False if raw_row_data_cost > 0 else True

        for cost_data_group in cost_data["Groups"]:
          raw_row_data_cost_group = Decimal(cost_data_group["Metrics"][cost_metric]["Amount"])
          row_data_cost_group = raw_row_data_cost_group

          if total_attribute_empty:
            raw_row_data_cost += raw_row_data_cost_group
          
          if year_data[cost_metric][by_month_key]["days"][day_key]["currency"] != cost_data_group["Metrics"][cost_metric]["Unit"]:
            row_data_cost_group = self.get_common().helper_currency().convert(
              ammount= row_data_cost_group,
              currency_from= cost_data_group["Metrics"][cost_metric]["Unit"],
              currency_to= year_data[cost_metric][by_month_key]["days"][day_key]["currency"],
              conversion_date= self.get_common().helper_type().datetime().yesterday(dt=self.get_common().helper_type().datetime().datetime_from_string(
                dt_string= self.get_common().helper_type().datetime().datetime_as_string(
                  dt= data_dt,
                  dt_format= "%Y%m01"
                ),
                dt_format= "%Y%m%d"
              )).date()
            )
          for cost_data_group_key in cost_data_group["Keys"]:
            if (year_data[cost_metric][by_month_key]["days"][day_key]["resource_type"][f'{total_key}'].get(cost_data_group_key) is None or
                year_data[cost_metric][by_month_key]["days"][day_key]["resource_type"][f'origional_currency_{total_key}'].get(cost_data_group_key) is None):
              year_data[cost_metric][by_month_key]["days"][day_key]["resource_type"][f'origional_currency_{total_key}'][cost_data_group_key] = Decimal(0)
              year_data[cost_metric][by_month_key]["days"][day_key]["resource_type"][f'{total_key}'][cost_data_group_key] = Decimal(0)
            
            if (year_data[cost_metric][by_month_key]["totals"]["resource_type"][f'{total_key}'].get(cost_data_group_key) is None or
                year_data[cost_metric][by_month_key]["totals"]["resource_type"][f'origional_currency_{total_key}'].get(cost_data_group_key) is None):
              year_data[cost_metric][by_month_key]["totals"]["resource_type"][f'origional_currency_{total_key}'][cost_data_group_key] = Decimal(0)
              year_data[cost_metric][by_month_key]["totals"]["resource_type"][f'{total_key}'][cost_data_group_key] = Decimal(0)

            year_data[cost_metric][by_month_key]["days"][day_key]["resource_type"][f'origional_currency_{total_key}'][cost_data_group_key] += Decimal(raw_row_data_cost_group)
            year_data[cost_metric][by_month_key]["days"][day_key]["resource_type"][f'{total_key}'][cost_data_group_key] += Decimal(row_data_cost_group)
            year_data[cost_metric][by_month_key]["totals"]["resource_type"][f'origional_currency_{total_key}'][cost_data_group_key] += Decimal(raw_row_data_cost_group)
            year_data[cost_metric][by_month_key]["totals"]["resource_type"][f'{total_key}'][cost_data_group_key] += Decimal(row_data_cost_group)
        
        row_data_cost = raw_row_data_cost
        
        if year_data[cost_metric][by_month_key]["days"][day_key]["currency"] != year_data[cost_metric][by_month_key]["days"][day_key]["origional_currency"]:
          row_data_cost = self.get_common().helper_currency().convert(
            ammount= row_data_cost,
            currency_from= year_data[cost_metric][by_month_key]["days"][day_key]["origional_currency"],
            currency_to= year_data[cost_metric][by_month_key]["days"][day_key]["currency"],
            conversion_date= self.get_common().helper_type().datetime().yesterday(dt=self.get_common().helper_type().datetime().datetime_from_string(
              dt_string= self.get_common().helper_type().datetime().datetime_as_string(
                dt= data_dt,
                dt_format= "%Y%m01"
              ),
              dt_format= "%Y%m%d"
            )).date()
          )

        year_data[cost_metric][by_month_key]["days"][day_key][f'origional_currency_{total_key}'] += Decimal(raw_row_data_cost)
        year_data[cost_metric][by_month_key]["days"][day_key][f'{total_key}'] += Decimal(row_data_cost)
        year_data[cost_metric][by_month_key]["totals"][f'{total_key}'] += Decimal(row_data_cost)
        if data_dt >= fiscal_start and data_dt <= fiscal_end:
          year_data[cost_metric][by_month_key]["totals"][f'fiscal_{total_key}'] += Decimal(row_data_cost)


  async def __process_get_cost_data(self, account, client, fiscal_year_start, loop, *args, **kwargs):
    fiscal_year_start_date = self.get_common().helper_type().datetime().datetime_from_string(
      dt_string= f"{self.get_data_start().year}/{fiscal_year_start}",
      dt_format= "%Y/%m/%d"
    )
    
    if fiscal_year_start_date > self.get_data_start():
      fiscal_year_start_date = fiscal_year_start_date + self.get_common().helper_type().datetime().time_delta(years= -1)

    fiscal_year_end = self.get_common().helper_type().datetime().yesterday(dt= (fiscal_year_start_date
                 + self.get_common().helper_type().datetime().time_delta(years= 1, dt= fiscal_year_start_date)))
    

    start_date = (fiscal_year_start_date
                 + self.get_common().helper_type().datetime().time_delta(months= -1, dt= fiscal_year_start_date))
    
    forecast_end = (fiscal_year_end
                 + self.get_common().helper_type().datetime().time_delta(months= 3, dt= fiscal_year_end))
    

    year_data = {}
    cost_metric = "NetUnblendedCost"
    await self.__process_get_cost_data_process_year_data(
      year_data= year_data,
      client= client,
      account= account,
      start_date= start_date,
      end_date= self.get_data_start() if forecast_end > self.get_data_start() else forecast_end,
      fiscal_start= fiscal_year_start_date, 
      fiscal_end= fiscal_year_end,
      cost_metrics = [cost_metric]
    )

    if forecast_end > self.get_data_start():
      await self.__process_get_cost_data_process_forcast(
        year_data= year_data,
        client= client,
        account= account,
        start_date= self.get_data_start(),
        end_date= forecast_end,
        fiscal_start= fiscal_year_start_date, 
        fiscal_end= fiscal_year_end,
        cost_metrics = [cost_metric]
      )

    month_key = self.get_common().helper_type().datetime().datetime_as_string(
      dt= self.get_data_start(),
      dt_format= "%Y%m"
    )
    last_month_key = self.get_common().helper_type().datetime().datetime_as_string(
      dt= (self.get_data_start() - self.get_common().helper_type().datetime().time_delta(days= (self.get_data_start().day + 1))),
      dt_format= "%Y%m"
    )
    
    return_data = {
      "cost_metric": cost_metric,
      "year_to_date": Decimal(0),  
      "year_forecast": Decimal(0),
      "fiscal_year_to_date": Decimal(0),  
      "fiscal_year_forecast": Decimal(0),
      "month_to_date": Decimal(0),  
      "month_forecast": Decimal(0),
      "last_seven_days": Decimal(0),
      "raw_last_14_days": {},
      "last_month": Decimal(0),
    }

    day_count = 0
    for i in range(0,14):     
      month_key_last14 = self.get_common().helper_type().datetime().datetime_as_string(
        dt= (self.get_data_start() - self.get_common().helper_type().datetime().time_delta(days= i)),
        dt_format= "%Y%m"
      )
      day_key = self.get_common().helper_type().datetime().datetime_as_string(
        dt= (self.get_data_start() - self.get_common().helper_type().datetime().time_delta(days= i)),
        dt_format= "%Y%m%d"
      )

      if year_data[cost_metric][month_key_last14]["days"].get(day_key) is None:
        continue
      
      return_data["raw_last_14_days"][day_key] = year_data[cost_metric][month_key_last14]["days"].get(day_key)

      if day_count >= 7:
        continue

      day_count += 1
      return_data["last_seven_days"] += year_data[cost_metric][month_key_last14]["days"][day_key]["total"]
      
    for data in year_data[cost_metric].values():
      return_data["fiscal_year_to_date"] += data["totals"].get("fiscal_total")
      return_data["fiscal_year_forecast"] += (data["totals"].get("fiscal_total") + data["totals"].get("fiscal_forcast_total"))
      if data["year"] == self.get_data_start().year:
        return_data["year_to_date"] += data["totals"].get("total")
        return_data["year_forecast"] += (data["totals"].get("total") + data["totals"].get("forcast_total"))

    
    if year_data[cost_metric].get(month_key) is not None:
      return_data["month_to_date"] = year_data[cost_metric][month_key]["totals"]["total"]
      return_data["month_forecast"] = year_data[cost_metric][month_key]["totals"]["total"] + year_data[cost_metric][month_key]["totals"]["forcast_total"]
    
    if year_data[cost_metric].get(last_month_key) is not None:
      return_data["last_month"] = year_data[cost_metric][last_month_key]["totals"]["total"]
  
    return return_data


  async def _process_account_data(self, account, loop, *args, **kwargs):
    if self.get_common().helper_type().string().is_null_or_whitespace(string_value= kwargs.get("fiscal_year_start")):
      kwargs["fiscal_year_start"] = self.get_cloud_data_client().get_default_fiscal_year_start()
    
    client = self.get_cloud_client().get_boto_client(client= 'ce',  account=account)
  
    return {
      "account": account,
      "data": [ self.get_common().helper_type().dictionary().merge_dictionary([
        {},
        await self.get_base_return_data(
          account= account,
          resource_id =  f'{self.get_cloud_client().get_account_prefix()}{self.get_cloud_client().get_account_id(account= account)}',
        ),
        await self.__process_get_cost_data(account= account, client= client, loop= loop, *args, **kwargs)
      ])]
    }

  



from threemystic_common.base_class.base_provider import base
from abc import abstractmethod
import asyncio, concurrent.futures, socket, errno
from tqdm.asyncio import tqdm

class cloud_data_client_provider_base_data(base):
  def __init__(self, *args, **kwargs):
    super().__init__(*args, **kwargs)
    
    self._set_cloud_data_client(*args, **kwargs)
    self._set_client_name(*args, **kwargs)
    self._set_max_process_pool(*args, **kwargs)
    self._set_max_thread_pool(*args, **kwargs)
    self._set_data_start(*args, **kwargs)

    self._process_preloaded_data(*args, **kwargs)

  
  @abstractmethod
  def get_accounts(self):
    pass
  
  @abstractmethod
  async def _process_account_data(self, *args, **kwargs):
    raise Exception("_process_account_data is not DEFINED")
  
  @property
  def required_extra_data(self, *args, **kwargs):
    if hasattr(self, "_required_extra_data"):
      return self._required_extra_data
    
    return {}
  
  @required_extra_data.setter
  def required_extra_data(self, value, *args, **kwargs):
    self._required_extra_data = value

  def _process_preloaded_data(self, preloaded_data = None, *args, **kwargs):
    pass
  
  def get_data_start(self, *args, **kwargs):
    return self.__data_start
  
  def _set_data_start(self, *args, **kwargs):
    self.__data_start = self.get_common().helper_type().datetime().get()

  def get_max_process_pool(self, *args, **kwargs):
    return self._max_process_pool
  
  def _set_max_process_pool(self, max_thread_pool = 5, *args, **kwargs):
    self._max_process_pool = max_thread_pool

  def get_max_thread_pool(self, *args, **kwargs):
    return self._max_thread_pool
  
  def _set_max_thread_pool(self, max_thread_pool = 35, *args, **kwargs):
    self._max_thread_pool = max_thread_pool
  
  def get_cloud_client(self, *args, **kwargs):
    return self._get_cloud_data_client_raw().get_cloud_client()
  
  def _get_cloud_data_client_raw(self, *args, **kwargs):
    return self._cloud_data_client
  
  def get_cloud_data_client(self, *args, **kwargs):
    return self._get_cloud_data_client_raw().get_cloud_data_client()
  
  def _set_cloud_data_client(self, cloud_data_client, *args, **kwargs):
    self._cloud_data_client = cloud_data_client

  def get_client_name(self, *args, **kwargs):
    return self._client_name
  
  def _set_client_name(self, data_action, *args, **kwargs):
    self._client_name = f"{self.get_provider()}-{data_action}-data"
  
  async def _pre_load_main_process(self, pool, *args, **kwargs):
    pass

  async def _get_environment(self, account = None, resource = None, *args, **kwargs):
    resource_name = self.get_common().helper_type().string().set_case(string_value= self.get_cloud_client().get_resource_name(resource= resource), case= "lower")
    account_name = self.get_common().helper_type().string().set_case(string_value= self.get_cloud_client().get_account_name(account= account), case= "lower")
    
    if not self.get_common().helper_type().string().is_null_or_whitespace(string_value= resource_name):
      for name in self.get_cloud_data_client().get_nonprod_names():
        if name in resource_name:
          return name
    
    for name in self.get_cloud_data_client().get_nonprod_names():
      if name in account_name:
        return name
      
    return "prod"
  
  async def get_base_return_data(self, account= None, resource_id = None, resource = None, region= None, resource_groups= None, resource_tags_resource = None, *args, **kwargs):
    
    resource_data = self.get_common().helper_type().dictionary().merge_dictionary([
      {},
      {
        "extra_account": self.get_common().helper_type().dictionary().merge_dictionary([
          {},
          {
            "extra_environment": await self._get_environment(account= account),
            "extra_tags": self.get_cloud_client().get_resource_tags_as_dictionary(resource= account)
          },
          self.get_cloud_client().serialize_resource(resource= account),
        ]),
        "extra_region": (
          region if not None else (
            self.get_cloud_client().get_resource_location(resource= resource) if resource is not None else None)),
        "extra_resourcegroups": resource_groups,
        "extra_environment": await self._get_environment(account= account, resource= resource),
        "extra_tags": self.get_cloud_client().get_resource_tags_as_dictionary(resource= (resource if resource_tags_resource is None else resource_tags_resource)),
        "extra_id": (
          resource_id if not None else (
            self.get_cloud_client().get_account_id(account= account) if account is not None else  None)),
      },
      self.get_cloud_client().serialize_resource(resource= resource) if resource is not None else self.get_cloud_client().serialize_resource(resource= account)
    ])

    resource_data["extra_resourcegroups"] = self.get_common().helper_type().list().unique_list(
      data= resource_data.get("extra_resourcegroups"),
      case_sensitive = False
    )
    return resource_data
  
  def format_results(self, results, output_format = None, *args, **kwargs):        
    if output_format is None or (self.get_common().helper_type().string().set_case(string_value= output_format, case= "lower") not in self.get_supported_output_format()):
      output_format = "json"
    
    try:
      if output_format == "yaml":
        print(self.get_common().helper_yaml().dumps(data= results))
        return
        # print(results)
        # work in progress
        # from tabulate import tabulate
        # tablulate_data = []
        # headers = None
        # for _, item_data in results.items():
        #   if len(item_data) > 0 and headers is None:
        #     headers = [key for key in item_data[0].keys()]
        #   for data in item_data:
        #     tablulate_data += [ (val if self.common.is_numeric(val) or self.common.is_type(val, str) else json.dumps(val, default=self.common.json_dumps_serializable_default))  for val in data.values() ] 
        # print(tablulate_data)
        # print (tabulate(
        #   tabular_data= tablulate_data, 
        #   headers= headers,
        # ))
      
      print( self.get_common().helper_json().dumps(data= results, indent=2))
      return
        
    
    
      
    except socket.error as e:
      if e.errno != errno.EPIPE:
        raise

  async def _process_get_data_by_id(self, results, id_override = None, *args, **kwargs):
    results_by_id = {}
    if self.get_common().helper_type().string().is_null_or_whitespace(string_value= id_override):
      id_override = "extra_id"
    for _, result_data in results.items():
      for data in result_data:
        results_by_id[data[id_override]] = data
    
    return results_by_id 
  
  async def get_data_by_id(self, results = None, id_override = None, *args, **kwargs):
    if results is None:
      results = await self.main( *args, **kwargs)
      
    return await self._process_get_data_by_id(results= results, id_override= id_override) 
  
  async def __main_poolexecutor(self, *args, **kwargs):   
    with concurrent.futures.ThreadPoolExecutor(self.get_max_thread_pool()) as pool:
        return await self.main_process(
          pool= pool,
          **kwargs
        )
  
  def get_runparam_key(self, data_key = None, default_value = {}, *args, **kwargs):
    if data_key is None:
      return self.__run_params
    
    if data_key == "send_account_data_lambda":
      return self.__run_params_send_account_data_lambda
    
    if data_key is not None and data_key in self.__run_params:
      if default_value is None or self.get_common().helper_type().general().is_type(obj= self.__run_params.get(data_key), type_check= type(default_value)):
        return self.__run_params.get(data_key)
    
    return default_value
  
  def _set_run_params(self, run_params = None, *args, **kwargs):
    if run_params is None:
      self.__run_params = {}
      return
    
    if not self.get_common().helper_type().general().is_type(obj= run_params, type_check= dict):
      self.__run_params = {}
      return
    
    if "data_action" in run_params:
      del run_params["data_action"]
    
    if "data_filter" in run_params:
      if not self.get_common().helper_type().general().is_type(obj= run_params.get("data_filter"), type_check= dict):
        run_params["data_filter"] = {}
    
    if "data_accounts" in run_params:
      if self.get_common().helper_type().general().is_type(obj= run_params.get("data_accounts"), type_check= str):
        run_params["data_accounts"] = self.get_common().helper_type().string().split(
          string_value= run_params["data_accounts"],
          separator= r"[,;\s]"
        )
      if not self.get_common().helper_type().general().is_type(obj= run_params.get("data_accounts"), type_check= list):
        run_params["data_accounts"] = []
    
    self.__run_params_send_account_data_lambda = None
    if "send_account_data_lambda" in run_params:
      self.__run_params_send_account_data_lambda = run_params.pop("send_account_data_lambda")

    self.__run_params = self.get_common().helper_type().dictionary().merge_dictionary([
      {},
      {
        "data_filter": {},
        "data_hideempty": False,
        "data_accounts": None,
      },
      run_params
    ])
    
  
  
  def _process_data_filter_condition_in(self, condition, condition_value, data_value, *args, **kwargs):
    condition_settings = self._process_data_filter_condition_settings("in", condition)

    if condition_settings.get("case_insensitive") is True:
      data_value = self.get_common().helper_type().string().set_case(string_value= data_value, case= "lower")
      if self.get_common().helper_type().general().is_type(obj= condition_value, type_check= str):
        condition_value = self.get_common().helper_type().string().set_case(string_value= condition_value, case= "lower")
      elif self.get_common().helper_type().general().is_type(obj= condition_value, type_check= list):
        condition_value = [self.get_common().helper_type().string().set_case(string_value= value, case= "lower") for value in condition_value]
      else:
        raise self.get_common().exception().exception(
          exception_type = "argument"
        ).not_implemented(
          logger= self.get_common().get_logger(), 
          name = "condition_value",
          message = f"condition_value should be either string or list: condition_value - {type(condition_value)}"
        )

    if condition_settings.get("not") is True:
      return data_value not in condition_value
    
    return data_value in condition_value
  
  def _process_data_filter_condition_less_than(self, condition, condition_value, data_value, *args, **kwargs):
    condition_settings = self._process_data_filter_condition_settings("lt", condition)

    if condition_settings.get("case_insensitive") is True:
      data_value = self.get_common().helper_type().string().set_case(string_value= data_value, case= "lower")
      condition_value = self.get_common().helper_type().string().set_case(string_value= condition_value, case= "lower")

    if condition_settings.get("not") is True and condition_settings.get("equals") is False:
      return not (data_value < condition_value)
    
    if condition_settings.get("not") is True and condition_settings.get("equals") is True:
      return not (data_value <= condition_value)
    
    if condition_settings.get("equals") is True:
      return (data_value <= condition_value)
    
    return data_value < condition_value
  
  def _process_data_filter_condition_greater_than(self, condition, condition_value, data_value, *args, **kwargs):
    condition_settings = self._process_data_filter_condition_settings("gt", condition)

    if condition_settings.get("case_insensitive") is True:
      data_value = self.get_common().helper_type().string().set_case(string_value= data_value, case= "lower")
      condition_value = self.get_common().helper_type().string().set_case(string_value= condition_value, case= "lower")

    if condition_settings.get("not") is True and condition_settings.get("equals") is False:
      return not (data_value > condition_value)
    
    if condition_settings.get("not") is True and condition_settings.get("equals") is True:
      return not (data_value >= condition_value)
    
    if condition_settings.get("equals") is True:
      return (data_value >= condition_value)
    
    return data_value > condition_value
  
  def _process_data_filter_condition_equals(self, condition, condition_value, data_value, *args, **kwargs):
    
    condition_settings = self._process_data_filter_condition_settings("equals", condition)

    if condition_settings.get("case_insensitive") is True:
      data_value = self.get_common().helper_type().string().set_case(string_value= data_value, case= "lower")
      condition_value = self.get_common().helper_type().string().set_case(string_value= condition_value, case= "lower")

    if condition_settings.get("not") is True:
      return data_value != condition_value
    
    return data_value == condition_value
    
  
  
  def _process_data_filter_condition_settings(self, condition, condition_value, *args, **kwargs):
    condition_settings = {
      "not": False,
      "case_insensitive": False,
      "equals": False
    }
    if condition == condition_value:
      return condition_settings
    
    while condition != condition_value:
      if condition_value[0:3] == "not":
        condition_settings["not"] = True
        condition_value=condition_value[3:]
        continue
      if condition_value[0:1] == "i":
        condition_settings["case_insensitive"] = True
        condition_value=condition_value[1:]
        continue
      if condition_value[0:1] == "e":
        condition_settings["equals"] = True
        condition_value=condition_value[1:]
        
      break

    return condition_settings
    

  def _process_data_filter_condition(self, condition, data_item, *args, **kwargs):
    if condition is None:
      return True
    
    if self.get_common().helper_type().string().is_null_or_whitespace(string_value= condition.get("condition")):
      return True

    key_value = condition.get("key")
    if key_value is None:
      return True
    if self.get_common().helper_type().general().is_type(obj= key_value, type_check= str):
      if self.get_common().helper_type().string().is_null_or_whitespace(string_value= key_value):
        return True

    condition_operator = self.get_common().helper_type().string().set_case(string_value= condition.get("condition"), case= "lower")
    data_value = self.get_common().helper_type().general().get_container_value(container= data_item, value_key= key_value)
    
    if condition_operator.endswith("equals"):
      return self._process_data_filter_condition_equals(condition= condition_operator, condition_value= condition.get("value"), data_value= data_value)
    
    if condition_operator.endswith("gt"):
      return self._process_data_filter_condition_greater_than(condition= condition_operator, condition_value= condition.get("value"), data_value= data_value)
    
    if condition_operator.endswith("lt"):
      return self._process_data_filter_condition_less_than(condition= condition_operator, condition_value= condition.get("value"), data_value= data_value)
    
    if condition_operator.endswith("in"):
      return self._process_data_filter_condition_in(condition= condition_operator, condition_value= condition.get("value"), data_value= data_value)
    
    if condition_operator.endswith("and"):
      if condition.get("filters") is None:
        return True
      if self.get_common().helper_type().general().is_type(obj= condition.get("filters"), type_check= dict):
        condition["filters"] = [condition.get("filters")]
      
      running_value = True
      for filter in condition.get("filters"):
        running_value = running_value and self._process_data_filter_condition(
          condition= filter,
          data_item= data_item
        )
      
      return running_value
    
    if condition_operator.endswith("or"):
      if condition.get("filters") is None:
        return True
      if self.get_common().helper_type().general().is_type(obj= condition.get("filters"), type_check= dict):
        condition["filters"] = [condition.get("filters")]
      
      running_value = False
      for filter in condition.get("filters"):
        running_value = running_value or self._process_data_filter_condition(
          condition= filter,
          data_item= data_item
        )
      
      return running_value


  def _process_data_filter_is_row_valid(self, data_item, *args, **kwargs):
    data_filter = self.get_runparam_key(data_key= "data_filter", default_value= {})
    if len(data_filter) < 1:
      return True
    
    return self._process_data_filter_condition(condition= data_filter, data_item= data_item)
    

  def _process_data_filter(self, data, *args, **kwargs):
    if len(self.get_runparam_key(data_key= "data_filter", default_value= {})) < 1:
      return data
    
    return [data_item for data_item in data if self._process_data_filter_is_row_valid(data_item= data_item)]


  async def main(self, pool= None, run_params = None, *args, **kwargs):   

    self._set_run_params(run_params= run_params)
    if pool == None:
      return await self.__main_poolexecutor(*args, **kwargs)
    
    return await self.main_process(
          pool= pool,
          **kwargs
        )
  
  async def _progressbar_close(self, progressbar):
    if progressbar is None:
      return
    
    await self._progressbar_update(progressbar)
    progressbar.close()
    await asyncio.sleep(.25)

  async def _progressbar_update(self, progressbar, update_count = None):
    if progressbar is None:
      return
    
    if update_count is None:
      progressbar.update()  
      return

    progressbar.update(n=update_count)

  async def _process_done_all_complete(self, pending_tasks, timeout = None, *args, **kwargs):
    done_tasks, pending_tasks = await asyncio.wait(pending_tasks, timeout= timeout, return_when= asyncio.ALL_COMPLETED)
    
    return {
      "done": [task_done for task_done in done_tasks],
      "tasks": [task_pending for task_pending in pending_tasks]
    } 

  async def _process_done_first_complete(self, pending_tasks, timeout = .25, *args, **kwargs):

    done_tasks, pending_tasks = await asyncio.wait(pending_tasks, timeout= timeout, return_when= asyncio.FIRST_COMPLETED)

    return {
      "done": [task_done for task_done in done_tasks],
      "tasks": [task_pending for task_pending in pending_tasks]
    }
  
  async def _process_done_auto(self, pending_tasks, timeout = .25, return_when = asyncio.FIRST_COMPLETED, *args, **kwargs):
    if return_when == asyncio.FIRST_COMPLETED:
      return await self._process_done_first_complete(pending_tasks= pending_tasks, timeout= timeout)
    
    return await self._process_done_all_complete(pending_tasks= pending_tasks, timeout= timeout)
  
  async def process_done(self, pending_tasks, account_progress = None, timeout = .25, return_when = asyncio.FIRST_COMPLETED, *args, **kwargs):
    if len(pending_tasks) < 1:
      return {
        "tasks": pending_tasks,
        "tasks_processed": [],
        "done_count": 0,
      }
    
    done_results = await self._process_done_auto(pending_tasks= pending_tasks, timeout= timeout, return_when= return_when)
    done_count = len(done_results["done"])

    if done_count > 0 and account_progress is not None:
      await self._progressbar_update(progressbar= account_progress, update_count= done_count)
      
    return {
        "tasks": done_results["tasks"],
        "tasks_processed": done_results["done"],
        "done_count": done_count,
    }
  
  async def _process_account(self, *args, **kwargs):
    return_data = await self._process_account_data(*args, **kwargs)
    if self.get_runparam_key(data_key= "send_account_data_lambda", default_value= None) is None:
      return return_data
    
    await self.get_runparam_key(data_key= "send_account_data_lambda", default_value= None)["handler"](data= {
      (self.get_runparam_key(data_key= "send_account_data_lambda", default_value= None)["sheet_key"]): {
        self.get_cloud_client().get_account_id(account= return_data["account"]): return_data.pop("data")
      }
    })
    

    return_data["data"] = []
    return return_data
  
  async def process_account(self, *args, **kwargs):
    return await self._process_account(*args, **kwargs)
  
  def process_account_sync(self, *args, **kwargs):
    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)
    
    results =  loop.run_until_complete(self.process_account(loop= loop, *args, **kwargs))
    loop.close()
    return results
  
  async def main_process(self, pool, loop= None, *args, **kwargs):    
    
    print(f"Running: {self.get_client_name()} - {self.get_data_start()}")

    await self._pre_load_main_process(pool= pool)

    accounts = self.get_accounts(*args, **kwargs)

    tasks = []
    tasks_processed = []
    if loop is None:
      loop = asyncio.get_event_loop()

    for account in accounts:
      tasks.append(loop.run_in_executor(pool, lambda item:self.process_account_sync(**item), {"account":account, "pool":pool}))

    finalized_results = await self.get_common().helper_parallel_processing().ensure_all_tasks_complete(
      done_function= self.process_done,
      done_function_params = {"pending_tasks": tasks, "account_progress": None, "timeout": None, "return_when": asyncio.ALL_COMPLETED, "loop": loop},
      total_tasks= len(accounts),
      current_running_total=0
    )

    for finalized_result in finalized_results["process_result_data"]:
      tasks_processed += (finalized_result["tasks_processed"])

    return_data = {}
    for task in tasks_processed:
      if task.result() is None or (task.result() is not None and len(task.result()) < 1):
        continue
      
      if task.result().get("account") is None:
        continue
      
      task.result()["data"]= self._process_data_filter(data= task.result()["data"])
      if self.get_runparam_key(data_key= "data_hideempty", default_value= False) is True and len(task.result()["data"]) < 1:
        continue

      
      if return_data.get(self.get_cloud_client().get_account_id(account= task.result()["account"])) is None:
        return_data[self.get_cloud_client().get_account_id(account= task.result()["account"])] = task.result()["data"]
        continue
      
      return_data[self.get_cloud_client().get_account_id(account= task.result()["account"])] += task.result()["data"]
      
    return return_data

  
  
    

  
  




from threemystic_cloud_data_client.cloud_providers.azure.config.base_class.base import cloud_data_client_azure_config_base as base
from threemystic_common.base_class.generate_data.generate_data_handlers import generate_data_handlers



class cloud_data_client_azure_config_step_1(base):
  def __init__(self, *args, **kwargs):
    super().__init__(logger_name= "cloud_data_client_azure_config_step_1", *args, **kwargs)
    

  def check_cloud_client(self, *args, **kwargs):
    response = self.get_common().generate_data().generate(
      generate_data_config = {
        "base_config": {
          "validation": lambda item: self.get_common().helper_type().bool().is_bool(check_value= item),
          "messages":{
            "validation": f"Valid options for Yes are: {self.get_common().helper_type().bool().is_true_values()}",
          },
          "conversion": lambda item: self.get_common().helper_type().bool().is_true(check_value= item),
          "desc": f"Do you need to configure the Cloud Client.\nValid Options: {self.get_common().helper_type().bool().is_true_values()}",
          "default": None,
          "handler": generate_data_handlers.get_handler(handler= "base"),
          "optional": True
        }
      }
    )

    if response is None:
      return
    
    if not self.get_common().helper_type().bool().is_true(check_value= response.get("base_config").get("formated")):
      return


  def step(self, *args, **kwargs):
    
    if not super().step(run_base_config= True):
      return
    
    print()
    print()
    print()
    print(f"No additional config is required at this time for Data Client: {self.get_provider()}")
    # if steps are added will need to add a complete step
    
    self.check_cloud_client(*args, **kwargs)
    

    response = self.get_common().generate_data().generate(
      generate_data_config = {
        "base_config": {
          "validation": lambda item: self.get_common().helper_type().bool().is_bool(check_value= item),
          "messages":{
            "validation": f"Valid options for Yes are: {self.get_common().helper_type().bool().is_true_values()}",
          },
          "conversion": lambda item: self.get_common().helper_type().bool().is_true(check_value= item),
          "desc": f"Data Client: Do you want to configure base config?\nLeave blank to exit.\nValid Options: {self.get_common().helper_type().bool().is_true_values()}",
          "default": None,
          "handler": generate_data_handlers.get_handler(handler= "base"),
          "optional": True
        }
      }
    )

    if response is None:
      return
    
    if not self.get_common().helper_type().bool().is_true(check_value= response.get("base_config").get("formated")):
      return
    
    from threemystic_cloud_data_client.cloud_providers.general  import cloud_data_client_general as client
    client(common= self.get_common()).action_config()
    
    
  



from threemystic_cloud_data_client.cloud_providers.azure.client.actions.base_class.base import cloud_data_client_azure_client_action_base as base
import asyncio
from azure.mgmt.privatedns import PrivateDnsManagementClient
from azure.mgmt.dns import DnsManagementClient
from azure.mgmt.resource import ResourceManagementClient



class cloud_data_client_azure_client_action(base):
  def __init__(self, *args, **kwargs):
    super().__init__(
      data_action="dns", 
      logger_name= "cloud_data_client_azure_client_action_dns",
      *args, **kwargs)
  
  async def __process_get_resources_public(self, client, account, *args, **kwargs):    
    try:      
      return { self.get_cloud_client().get_resource_id(resource= dns):dns for dns in self.get_cloud_client().sdk_request(
          tenant= self.get_cloud_client().get_tenant_id(tenant= account, is_account= True), 
          lambda_sdk_command=lambda: client.zones.list()
        )
      }
           
    except Exception as err:
      self.get_common().get_logger().exception(
        msg= f"__process_get_resources_public: {err}",
        extra={
          "exception": err
        }
      )
      return {}
  
  async def __process_get_resources_public_record_sets(self, client, account, dns, *args, **kwargs):    
    try:      
      return [ record_set for record_set in self.get_cloud_client().sdk_request(
          tenant= self.get_cloud_client().get_tenant_id(tenant= account, is_account= True), 
          lambda_sdk_command=lambda: client.record_sets.list_all_by_dns_zone(resource_group_name= self.get_cloud_client().get_resource_group_from_resource(resource= dns), zone_name= dns.name)
        )
      ]
           
    except Exception as err:
      self.get_common().get_logger().exception(
        msg= f"__process_get_resources_public_record_sets: {err}",
        extra={
          "exception": err
        }
      )
      return []
  
  async def __process_get_resources_private_dns(self, client, account, *args, **kwargs):    
    try:      
      return { self.get_cloud_client().get_resource_id(resource= dns):dns for dns in self.get_cloud_client().sdk_request(
          tenant= self.get_cloud_client().get_tenant_id(tenant= account, is_account= True), 
          lambda_sdk_command=lambda: client.private_zones.list()
        )
      }
           
    except Exception as err:
      self.get_common().get_logger().exception(
        msg= f"__process_get_resources_private_dns: {err}",
        extra={
          "exception": err
        }
      )
      return {}
  
  async def __process_get_resources_private_dns_network(self, client, account, dns, *args, **kwargs):    
    try:      
      return [ network for network in self.get_cloud_client().sdk_request(
          tenant= self.get_cloud_client().get_tenant_id(tenant= account, is_account= True), 
          lambda_sdk_command=lambda: client.virtual_network_links.list(resource_group_name= self.get_cloud_client().get_resource_group_from_resource(resource= dns), private_zone_name= dns.name)
        )
      ]
           
    except Exception as err:
      self.get_common().get_logger().exception(
        msg= f"__process_get_resources_private_dns_network: {err}",
        extra={
          "exception": err
        }
      )
      return []
  
  async def __process_get_resources_private_dns_record_sets(self, client, account, dns, *args, **kwargs):    
    try:      
      return [ record_set for record_set in self.get_cloud_client().sdk_request(
          tenant= self.get_cloud_client().get_tenant_id(tenant= account, is_account= True), 
          lambda_sdk_command=lambda: client.record_sets.list(resource_group_name= self.get_cloud_client().get_resource_group_from_resource(resource= dns), private_zone_name= dns.name)
        )
      ]
           
    except Exception as err:
      self.get_common().get_logger().exception(
        msg= f"__process_get_resources_private_dns_record_sets: {err}",
        extra={
          "exception": err
        }
      )
      return []
  
  async def __process_get_resources_resource_dns(self, account, *args, **kwargs):    
    resource_client = ResourceManagementClient(credential= self.get_cloud_client().get_tenant_credential(tenant= self.get_cloud_client().get_tenant_id(tenant= account, is_account= True)), subscription_id= self.get_cloud_client().get_account_id(account= account))
    try:
        return {self.get_cloud_client().get_resource_id(resource= resource): resource for resource in self.get_cloud_client().sdk_request(
           tenant= self.get_cloud_client().get_tenant_id(tenant= account, is_account= True), 
           lambda_sdk_command=lambda: resource_client.resources.list(filter="resourceType eq 'Microsoft.Network/privateDnsZones' or resourceType eq 'Microsoft.Network/dnsZones'", expand="createdTime,changedTime,provisioningState")
          )
        }
    except Exception as err:
      self.get_common().get_logger().exception(
        msg= f"__process_get_resources_resource_dns: {err}",
        extra={
          "exception": err
        }
      )
      return {}
  
  async def __process_get_resources_dns_get(self, client_private, client_public, account, dns, value, *args, **kwargs):    
    try:

      if value is not None:
        return value
      
      if self.get_common().helper_type().string().set_case(string_value= dns.type, case= "lower") == "microsoft.network/privatednszones":
        return self.get_cloud_client().sdk_request(
            tenant= self.get_cloud_client().get_tenant_id(tenant= account, is_account= True), 
            lambda_sdk_command=lambda: client_private.private_zones.get(resource_group_name= self.get_cloud_client().get_resource_group_from_resource(resource= dns), private_zone_name= dns.name)
          )
      
      return self.get_cloud_client().sdk_request(
            tenant= self.get_cloud_client().get_tenant_id(tenant= account, is_account= True), 
            lambda_sdk_command=lambda: client_public.zones.get(resource_group_name= self.get_cloud_client().get_resource_group_from_resource(resource= dns), private_zone_name= dns.name)
          )
           
    except Exception as err:
      self.get_common().get_logger().exception(
        msg= f"__process_get_resources_dns_get: {err}",
        extra={
          "exception": err
        }
      )
      return dns
    
  
  async def _process_account_data(self, account, loop, *args, **kwargs):

    privatedns_client = PrivateDnsManagementClient(credential= self.get_cloud_client().get_tenant_credential(tenant= self.get_cloud_client().get_tenant_id(tenant= account, is_account= True)), subscription_id= self.get_cloud_client().get_account_id(account= account))
    publicdns_client = DnsManagementClient(credential= self.get_cloud_client().get_tenant_credential(tenant= self.get_cloud_client().get_tenant_id(tenant= account, is_account= True)), subscription_id= self.get_cloud_client().get_account_id(account= account))
    
    tasks = {
      "resources": loop.create_task(self.__process_get_resources_resource_dns(account= account)),
      "privatedns": loop.create_task(self.__process_get_resources_private_dns(client= privatedns_client, account= account)),
      "public": loop.create_task(self.__process_get_resources_public(client= publicdns_client, account= account))
    }

    await asyncio.wait(tasks.values())
    tasks_data = {}

    for id, dns in tasks["resources"].result().items():
      resource_type = self.get_common().helper_type().string().set_case(string_value= dns.type, case= "lower")
      
      tasks_data[f'{id};item'] = loop.create_task(self.__process_get_resources_dns_get(
        client_private= privatedns_client,
        client_public= publicdns_client,
        account= account,
        dns= dns,
        value= tasks["privatedns"].result().get(id) if resource_type == "microsoft.network/privatednszones" else tasks["public"].result().get(id)
      ))
      
      if resource_type == "microsoft.network/privatednszones":
        tasks_data[f'{id};record_sets'] = loop.create_task(self.__process_get_resources_private_dns_record_sets(
          client= privatedns_client,
          account= account,
          dns= dns
        ))
        tasks_data[f'{id};network'] = loop.create_task(self.__process_get_resources_private_dns_network(
          client= privatedns_client,
          account= account,
          dns= dns
        ))
        continue
        
      tasks_data[f'{id};record_sets'] = loop.create_task(self.__process_get_resources_public_record_sets(
        client= publicdns_client,
        account= account,
        dns= dns
      ))
    
    if len(tasks_data) > 0:
      await asyncio.wait(tasks_data.values())

    return {
      "account": account,
      "data": [
          self.get_common().helper_type().dictionary().merge_dictionary([
          {},
          await self.get_base_return_data(
            account= self.get_cloud_client().serialize_resource(resource= account),
            resource_id= self.get_cloud_client().get_resource_id(resource= tasks_data[f'{id};item'].result()),
            resource= tasks_data[f'{id};item'].result(),
            region= self.get_cloud_client().get_resource_location(resource= tasks_data[f'{id};item'].result()),
            resource_groups= [self.get_cloud_client().get_resource_group_from_resource(resource= tasks_data[f'{id};item'].result())],
          ),
          {
            "extra_record_sets": [ self.get_cloud_client().serialize_resource(resource= zone) for zone in tasks_data[f'{id};record_sets'].result() ] if tasks_data[f'{id};record_sets'].result() is not None else None,
            "extra_vnet_link": [ self.get_cloud_client().serialize_resource(resource= link) for link in tasks_data[f'{id};network'].result() ] if self.get_common().helper_type().string().set_case(string_value= raw_resource.type, case= "lower") == "microsoft.network/privatednszones" and tasks_data[f'{id};network'].result() is not None else None,
          },
        ]) for id, raw_resource in tasks["resources"].result().items()
      ]
    }


from threemystic_cloud_data_client.cloud_providers.azure.client.actions.base_class.base import cloud_data_client_azure_client_action_base as base
import asyncio
from azure.mgmt.storage import StorageManagementClient
from azure.mgmt.monitor import MonitorManagementClient
from decimal import Decimal, ROUND_HALF_UP

class cloud_data_client_azure_client_action(base):
  def __init__(self, *args, **kwargs):
    super().__init__(
      data_action="cloudstorage", 
      logger_name= "cloud_data_client_azure_client_action_cloudstorage",
      *args, **kwargs)
  
  
  async def _process_account_data_storage_size(self, client, account, storage_account, **kwargs):
      try:
        storage_account_id= self.get_cloud_client().get_resource_id(resource= storage_account)
        storage_size = self.get_cloud_client().sdk_request(
          tenant= self.get_cloud_client().get_tenant_id(tenant= account, is_account= True), 
          lambda_sdk_command=lambda: client.metrics.list(
            resource_uri= storage_account_id,
            timespan= f'{self.get_common().helper_type().datetime().get_iso_datetime(dt= (self.get_data_start() - self.get_common().helper_type().datetime().time_delta(days= 1)))}/{self.get_common().helper_type().datetime().get_iso_datetime(dt= self.get_data_start())}',
            interval= "PT1H",
            metricnames= "UsedCapacity"
          )
        )
        interval_count = 0
        sum_storage = Decimal(0.0)
        
        for interval_value in storage_size.value:
          for ts in interval_value.timeseries:
            for data in ts.data:
              if data.average is not None:
                sum_storage += Decimal(data.average)
              interval_count += 1
        
        return Decimal(sum_storage/interval_count).quantize(Decimal('0'), ROUND_HALF_UP)
      except Exception as err:
        return None
        
  async def _process_account_data_blob_containers(self, client:StorageManagementClient, account, storage_account, **kwargs):
      try:
        resource_group= self.get_cloud_client().get_resource_group_from_resource(resource= storage_account)
        storage_account_name= self.get_cloud_client().get_resource_name_from_resource(resource= storage_account)
        return [self.get_cloud_client().serialize_resource(resource= item) for item in self.get_cloud_client().sdk_request(
          tenant= self.get_cloud_client().get_tenant_id(tenant= account, is_account= True), 
          lambda_sdk_command=lambda: client.blob_containers.list(resource_group_name= resource_group, account_name= storage_account_name)
        )]
      except:
        return []
      
  async def _process_storage_account_by_page(self, client:StorageManagementClient, account, **kwargs):
      
      process_object = []
      for _, page in self.get_cloud_client().sdk_request(
          tenant= self.get_cloud_client().get_tenant_id(tenant= account, is_account= True), 
          lambda_sdk_command=lambda: enumerate(client.storage_accounts.list().by_page())
        ):
        for storage_account in page:
          for blob_container in await self._process_account_data_blob_containers(client= client, account= account, storage_account= storage_account):
            process_object.append({
              "container": blob_container,
              "storage_account": storage_account
            })
          process_object.append({
            "container": None,
            "storage_account": storage_account
          })
      
      return process_object

  async def _process_account_data(self, account, loop, **kwargs):

    client = StorageManagementClient(credential= self.get_cloud_client().get_tenant_credential(tenant= self.get_cloud_client().get_tenant_id(tenant= account, is_account= True)), subscription_id= self.get_cloud_client().get_account_id(account= account))
    monitor_client = MonitorManagementClient(credential= self.get_cloud_client().get_tenant_credential(tenant= self.get_cloud_client().get_tenant_id(tenant= account, is_account= True)), subscription_id= self.get_cloud_client().get_account_id(account= account))
    process_object = await self._process_storage_account_by_page(client= client, account= account)
      
    return {
        "account": account,
        "data": [ self.get_common().helper_type().dictionary().merge_dictionary([
          {},
          await self.get_base_return_data(
            account= self.get_cloud_client().serialize_resource(resource= account),
            resource_id= self.get_cloud_client().get_resource_id(resource= item.get("container") if item.get("container") is not None else item.get("storage_account")),
            resource = item.get("container"),
            region= self.get_cloud_client().get_resource_location(resource= item.get("storage_account")),
            resource_groups= [self.get_cloud_client().get_resource_group_from_resource(resource= item.get("storage_account"))],
          ),
          {
            "extra_storage_account": item.get("storage_account")
          },
          { "extra_storageaccount_bytes_24hours": await self._process_account_data_storage_size(
              client= monitor_client,
              account= account,
              storage_account= item.get("storage_account")
            ) 
          } if item.get("container") is None else {}
        ]) for item in process_object]
    }


from threemystic_cloud_data_client.cli.actions.base_class.base import cloud_data_client_action_base as base


class cloud_data_client_config(base):
  def __init__(self, *args, **kwargs):
    super().__init__(*args, **kwargs)


  def _process_provider_aws(self, *args, **kwargs):
      from threemystic_cloud_data_client.cloud_providers.aws  import cloud_data_client_aws as client
      client(common= self._cloud_data_client.get_common()).action_config()


  def _process_provider_azure(self, *args, **kwargs):
      from threemystic_cloud_data_client.cloud_providers.azure import cloud_data_client_azure as client
      client(common= self._cloud_data_client.get_common()).action_config()

      

  



from threemystic_cloud_data_client.cloud_providers.base_class.base_client import cloud_data_client_provider_base_client as base

class cloud_data_client_azure_client(base):
  def __init__(self, *args, **kwargs):
    super().__init__(provider= "azure", logger_name= "cloud_data_client_azure_client", *args, **kwargs)

  def get_client_parser_args_actions(self, *args, **kwargs):
    return {}
    
    
  
  


import sys


def main(*args, **kwargs):
  from threemystic_cloud_data_client.cli import cloud_data_client_cli
  cloud_data_client_cli().main(*args, **kwargs)
  

if __name__ == '__main__':   
  main(sys.argv[1:])


from threemystic_cloud_data_client.cloud_providers.azure.client.actions.base_class.base import cloud_data_client_azure_client_action_base as base
import asyncio
from azure.mgmt.cosmosdb import CosmosDBManagementClient

class cloud_data_client_azure_client_action(base):
  def __init__(self, *args, **kwargs):
    super().__init__(
      data_action="clouddb", 
      logger_name= "cloud_data_client_azure_client_action_clouddb",
      *args, **kwargs)
  
  async def __process_get_db_cosmosdb(self, client, account, *args, **kwargs):    
    try:
      # I need to look into this more but the cassandra clusters list seems to be standalone
      # client.cassandra_clusters.list_by_subscription
      return_data = []
      for db in self.get_cloud_client().sdk_request(
        tenant= self.get_cloud_client().get_tenant_id(tenant= account, is_account= True), 
        lambda_sdk_command=lambda: client.database_accounts.list()
        ):
        return_data.append(
          {"extra":{
            "extra_dbtype": "cosmosdb"
          }, "resource": db 
        })
      
      return return_data
    except Exception as err:
      self.get_common().get_logger().exception(
        msg= f"__process_get_db_cosmos: {err}",
        extra={
          "exception": err
        }
      )
      return []
    
  async def _process_account_data(self, account, loop, *args, **kwargs):
    # database categories
    # https://azure.microsoft.com/en-us/products/category/databases/

    client = CosmosDBManagementClient(credential= self.get_cloud_client().get_tenant_credential(tenant= self.get_cloud_client().get_tenant_id(tenant= account, is_account= True)), subscription_id= self.get_cloud_client().get_account_id(account= account))


    tasks = {
       "main": loop.create_task(self.__process_get_db_cosmosdb(client= client,account= account)),
    }

    await asyncio.wait(tasks.values())
    
    return {
      "account": account,
      "data": [
          self.get_common().helper_type().dictionary().merge_dictionary([
          {},
          await self.get_base_return_data(
            account= self.get_cloud_client().serialize_resource(resource= account),
            resource_id= self.get_cloud_client().get_resource_id(resource= item.get("resource")),
            resource=  item.get("resource"),
            region= self.get_cloud_client().get_resource_location(resource=  item.get("resource")),
            resource_groups= [self.get_cloud_client().get_resource_group_from_resource(resource=  item.get("resource"))],
          ),
          item.get("extra"),
        ]) for item in tasks["main"].result()
      ]
    }



from threemystic_cloud_data_client.cloud_providers.azure.client.actions.base_class.base import cloud_data_client_azure_client_action_base as base
import asyncio
from azure.mgmt.synapse import SynapseManagementClient


class cloud_data_client_azure_client_action(base):
  def __init__(self, *args, **kwargs):
    super().__init__(
      data_action="datawarehouse", 
      logger_name= "cloud_data_client_azure_client_action_datawarehouse",
      *args, **kwargs)
  
 
  async def _process_account_data(self, account, loop, *args, **kwargs):
    
    client = SynapseManagementClient(credential= self.get_cloud_client().get_tenant_credential(tenant= self.get_cloud_client().get_tenant_id(tenant= account, is_account= True)), subscription_id= self.get_cloud_client().get_account_id(account= account))
    
    return {
        "account": account,
        "data": [
           self.get_common().helper_type().dictionary().merge_dictionary([
            {},
            await self.get_base_return_data(
              account= self.get_cloud_client().serialize_resource(resource= account),
              resource_id= self.get_cloud_client().get_resource_id(resource= item),
              resource= item,
              region= self.get_cloud_client().get_resource_location(resource= item),
              resource_groups= [self.get_cloud_client().get_resource_group_from_resource(resource= item)],
            ),
            {
              "extra_sql_pools": [
                self.get_cloud_client().serialize_resource(resource= pool)
                for pool in client.sql_pools.list_by_workspace(resource_group_name= self.get_cloud_client().get_resource_group_from_resource(resource= item), workspace_name= item.name)
              ],
              "extra_big_data_pools": [
                self.get_cloud_client().serialize_resource(resource= pool)
                for pool in client.big_data_pools.list_by_workspace(resource_group_name= self.get_cloud_client().get_resource_group_from_resource(resource= item), workspace_name= item.name)
              ] 
            },
          ]) for item in self.get_cloud_client().sdk_request(
           tenant= self.get_cloud_client().get_tenant_id(tenant= account, is_account= True), 
           lambda_sdk_command=lambda: client.workspaces.list()
          )
        ]
    }


from threemystic_cloud_data_client.cloud_providers.general.config.base_class.base import cloud_data_client_general_config_base as base
from threemystic_common.base_class.generate_data.generate_data_handlers import generate_data_handlers



class cloud_data_client_general_config_step_2(base):
  def __init__(self, *args, **kwargs):
    super().__init__(logger_name= "cloud_data_client_general_config_step_2", *args, **kwargs)
    

  def step(self, *args, **kwargs):
    if not super().step(run_base_config= True):
      return
    
    self.update_general_config_completed(status= True)
    
    print()
    print()
    print()
    print("-------------------------------------")
    print("Resource Environment Info")
    print("By default the system will mark all accounts/subscriptions/resources as nonprod/prod based on the name. If the tag exists the tag will override general prod/nonprod setting.")
    print("It will first check for a match on the resources if nothing is found it will then look to the account/subscription. It Defaults to prod, because if something cannot be identified as nonprod it should be treated as prod for saftey.")
    print("The default nonprod names are:")
    print(self.get_nonprod_names())
    print("-------------------------------------")
    print()
    print()
    print()
    if not self.has_tag_data_config():
      return self.step_tag_new()
    
    return step_tag_clear()
  
  def step_tag_new(self, *args, **kwargs):
    response = self.get_common().generate_data().generate(
      generate_data_config = {
        "environment_tag": {
          "validation": lambda item: self.get_common().helper_type().bool().is_bool(check_value= item),
          "messages":{
            "validation": f"Valid options for Yes are: {self.get_common().helper_type().bool().is_true_values()}",
          },
          "conversion": lambda item: self.get_common().helper_type().bool().is_true(check_value= item),
          "desc": f"Do you use tags to determine environment?\nValid Options: {self.get_common().helper_type().bool().is_true_values()}",
          "default": None,
          "handler": generate_data_handlers.get_handler(handler= "base"),
          "optional": True
        }
      }
    )
    if response is None:
      return

    if self.get_common().helper_type().bool().is_true(check_value= response.get("environment_tag").get("formated")):
      self.step_tag() 
    
    return
  
  def step_tag_clear(self, confirm = False, *args, **kwargs):

    desc = ("Do you want to reset and clear the environment tag data?" if confirm is False else
      "Are you sure you want to reset the environment tag data?")
    response = self.get_common().generate_data().generate(
      generate_data_config = {
        "environment_tag_clear": {
          "validation": lambda item: self.get_common().helper_type().bool().is_bool(check_value= item),
          "messages":{
            "validation": f"Valid options for Yes are: {self.get_common().helper_type().bool().is_true_values()}",
          },
          "conversion": lambda item: self.get_common().helper_type().bool().is_true(check_value= item),
          "desc": f"{desc}\nValid Options: {self.get_common().helper_type().bool().is_true_values()}",
          "default": None,
          "handler": generate_data_handlers.get_handler(handler= "base"),
          "optional": True
        }
      }
    )
    if response is None:
      return

    if confirm is False and  self.get_common().helper_type().bool().is_true(check_value= response.get("environment_tag_clear").get("formated")):
      return self.step_tag_clear(confirm= True) 
    
    if confirm is True and  self.get_common().helper_type().bool().is_true(check_value= response.get("environment_tag_clear").get("formated")):
      print("-----------------------------")
      print()
      print()
      print("Base Environment Tag cleared")
      print()
      print()
      print("-----------------------------")
      self.reset_config_environment_data()
      return self.step()

    print("-----------------------------")
    print()
    print()
    print("Base Environment Tag NOT cleared")
    print()
    print()
    print("-----------------------------")
    return
  
  def step_tag(self, *args, **kwargs):
    response = self.get_common().generate_data().generate(
      generate_data_config = {
        "tag": {
          "validation": lambda item: not self.get_common().helper_type().string().is_null_or_whitespace(string_value= item),
          "messages":{
            "validation": f"The value cannot be empty",
          },
          "conversion": lambda item: self.get_common().helper_type().string().trim(string_value= item),
          "desc": f"What is the main tag to determin the environment?",
          "default": self.get_environment_data_config_value("tag", None),
          "handler": generate_data_handlers.get_handler(handler= "base"),
          "optional": self.get_environment_data_config_value("tag", None) is not None
        },        
        "tag_insensitive": {
          "validation": lambda item: self.get_common().helper_type().bool().is_bool(check_value= item),
          "messages":{
            "validation": f"Valid options for Yes are: {self.get_common().helper_type().bool().is_true_values()}",
          },
          "conversion": lambda item: self.get_common().helper_type().bool().is_true(check_value= item),
          "desc": f"Treat the tags as case insensitive.\nIE EnvIronmeNt and environment would be the same.\nValid Options: {self.get_common().helper_type().bool().is_true_values()}",
          "default": self.get_environment_data_config_value("tag_insensitive", True),
          "handler": generate_data_handlers.get_handler(handler= "base"),
          "optional": True
        }
      }
    )
    if response is None:
      print("-----------------------------")
      print()
      print()
      print("Base Environment Tag NOT updated/Configured")
      print()
      print()
      print("-----------------------------")
      return
    for key, item in response.items():
      self._update_config_environment_data(config_key= key, config_value= item.get("formated"))
    self._save_config()

    print("-----------------------------")
    print()
    print()
    print("Base Environment Tag updated/configured")
    print()
    print()
    print("-----------------------------")
    
    if len(self.get_environment_data_config_value("environment_tag", [])) < 1:
      return self.step_add_alttag()
    
    return self.step_tag_edit_new()
  
  def step_tag_edit_new(self, *args, **kwargs):
    print("-----------------------------")
    print()
    print()
    print("Do you want to edit an existing alt tag or add new alt tags?")
    print()
    print("1: Edit Existing")
    print("2: Remove Existing")
    print("3: New Alt Tag")
    print()
    print()
    print("-----------------------------")
    response = self.get_common().generate_data().generate(
      generate_data_config = {
        "environment_tag_options": {
          "validation": lambda item: not self.get_common().helper_type().string().is_null_or_whitespace(string_value= item) and self.get_common().helper_type().general().is_integer(item) and self.get_common().helper_type().int().get(item) > 0 and self.get_common().helper_type().int().get(item) < 4,
          "messages":{
            "validation": f"Valid options are 1 - 3.",
          },
          "conversion": lambda item: self.get_common().helper_type().int().get(item) if not self.get_common().helper_type().string().is_null_or_whitespace(string_value= item) else None,
          "desc": f"Please select if you want to edit or add a new alt tag.\nValid Options: 1 - 3",
          "default": None,
          "handler": generate_data_handlers.get_handler(handler= "base"),
          "optional": False
        }
      }
    )
    if response is None:
      return

    if self.get_common().helper_type().int().get(response.get("environment_tag_options").get("formated")) == 1:
      return self.step_edit_remove__alttag(is_edit = True)
    
    if self.get_common().helper_type().int().get(response.get("environment_tag_options").get("formated")) == 1:
      return self.step_edit_remove__alttag(is_edit = False)
    
    if self.get_common().helper_type().int().get(response.get("environment_tag_options").get("formated")) == 3:
      return self.step_add_alttag() 
    
    return
  
  def step_edit_remove__alttag(self, is_edit = True, *args, **kwargs):
    if len(self.get_environment_data_config_value("environment_tag", [])) < 1:
      return self.step_add_alttag()
    
    print("-----------------------------")
    print()
    print()
    if is_edit:
      print("Which tag do you want to edit?")
    else:
      print("Which tag do you want to remove?")
    print()
    print("0: Exit")
    for index, val in enumerate(self.get_environment_data_config_value("environment_tag", [])):
      print(f"{index + 1}: {val}")
    print()
    print()
    print("-----------------------------")

    response = self.get_common().generate_data().generate(
      generate_data_config = {
        "alttag": {
          "validation": lambda item: not self.get_common().helper_type().string().is_null_or_whitespace(string_value= item) and self.get_common().helper_type().general().is_integer(item) and self.get_common().helper_type().int().get(item) >= 0 and self.get_common().helper_type().int().get(item) <= (len(drive_item_id_options[drive_item_position]) - 1),
          "messages":{
            "validation": f"Valid options are: 0 - {len(self.get_environment_data_config_value('environment_tag', []))}",
          },
          "conversion": lambda item: self.get_common().helper_type().int().get(item) if not self.get_common().helper_type().string().is_null_or_whitespace(string_value= item) else None,
          "desc": f"Please select the tenant to use \nValid Options: 0 -  {len(self.get_environment_data_config_value('environment_tag', []))}",
          "default": None,
          "handler": generate_data_handlers.get_handler(handler= "base"),
          "optional": True
        },
      }
    )

    if response is None:
      return self.step_tag_edit_new()
    
    if self.get_common().helper_type().int().get(response.get("alttag").get("formated")) == 0:
      return
    
    if not is_edit:
      self.get_environment_data_config_value("environment_tag", []).pop(self.get_common().helper_type().int().get(response.get("alttag").get("formated")) - 1)
      self._save_config_environment_data()
      return self.step_tag_edit_new()
    
    self.get_environment_data_config_value("environment_tag", [])[self.get_common().helper_type().int().get(response.get("alttag").get("formated")) - 1] = self.step_add_edit_alttag(edit_tag_index= self.get_common().helper_type().int().get(response.get("alttag").get("formated")) - 1)
    self._save_config_environment_data()
    return self.step_tag_edit_new()
  
  def step_add_edit_alttag(self, edit_tag_index = None, *args, **kwargs):
    
    default_tag_value = None
    if edit_tag_index is not None:
      if edit_tag_index >=0 and edit_tag_index <= len(self.get_environment_data_config_value("environment_tag", [])):
        default_tag_value = self.get_environment_data_config_value("environment_tag", [])[edit_tag_index]

    response = self.get_common().generate_data().generate(
      generate_data_config = {
        "environment_tag": {
          "validation": lambda item: not self.get_common().helper_type().string().is_null_or_whitespace(string_value= item),
          "messages":{
            "validation": f"The value cannot be empty",
          },
          "conversion": lambda item: self.get_common().helper_type().string().trim(string_value= item),
          "desc": f"What is the alt tag?\n(leave empty or type quit or exit)",
          "default": default_tag_value,
          "handler": generate_data_handlers.get_handler(handler= "base"),
          "optional": False
        }
      }
    )

    if response is None:
      return default_tag_value

    if response.get("environment_tag") is None:
      return default_tag_value

    if self.get_common().helper_type().string().is_null_or_whitespace(string_value= response.get("environment_tag").get("formated")):
      return default_tag_value
    
    return response.get("environment_tag").get("formated")
    


  def step_add_alttag(self, *args, **kwargs):
    
    stop_alt_tag = False
    alt_tags_add = []
    while not stop_alt_tag:
      response = self.get_common().generate_data().generate(
        generate_data_config = {
          "environment_tag": {
            "validation": lambda item: not self.get_common().helper_type().string().is_null_or_whitespace(string_value= item),
            "messages":{
              "validation": f"The value cannot be empty",
            },
            "conversion": lambda item: self.get_common().helper_type().string().trim(string_value= item),
            "desc": f"What is the alt tag?\n(leave empty to end or type quit or exit)",
            "default": None,
            "handler": generate_data_handlers.get_handler(handler= "base"),
            "optional": False
          }
        }
      )

      if response is None:
        stop_alt_tag = True
        break

      if response.get("environment_tag") is None:
        stop_alt_tag = True
        break

      if self.get_common().helper_type().string().is_null_or_whitespace(string_value= response.get("environment_tag").get("formated")):      
        stop_alt_tag = True
        break
    
      alt_tags_add.append(response.get("environment_tag").get("formated"))

    alt_tags_add = self.get_common().helper_type().list().unique_list(
      data= alt_tags_add,
      case_sensitive= not self.get_environment_data_config_value(
        config_key= "tag_insensitive",
        default_if_none= True
      )
    )
    self._update_config_environment_data(config_key= "alt_tags", config_value= alt_tags_add)
    self._save_config()
    
  



from threemystic_cloud_data_client.cloud_providers.aws.base_class.base import cloud_data_client_provider_aws_base as base
from threemystic_common.base_class.generate_data.generate_data_handlers import generate_data_handlers

class cloud_data_client_aws_config_base(base):
  def __init__(self, *args, **kwargs):
    super().__init__(*args, **kwargs)

  def step(self, *args, **kwargs):    
    return self.ensure_cloud_client_config_completed()


from threemystic_cloud_data_client.cloud_providers.azure.client.actions.base_class.base import cloud_data_client_azure_client_action_base as base
import asyncio
from azure.mgmt.compute import ComputeManagementClient
from azure.mgmt.resource import ResourceManagementClient


class cloud_data_client_azure_client_action(base):
  def __init__(self, *args, **kwargs):
    super().__init__(
      data_action="vmss", 
      logger_name= "cloud_data_client_azure_client_action_vmss",  
      *args, **kwargs)
  
  
    
  async def __process_get_resources_vmss(self, account, *args, **kwargs):
    resource_client = ResourceManagementClient(credential= self.get_cloud_client().get_tenant_credential(tenant= self.get_cloud_client().get_tenant_id(tenant= account, is_account= True)), subscription_id= self.get_cloud_client().get_account_id(account= account))
    try:
      return { self.get_cloud_client().get_resource_id(resource= resource): resource for resource in self.get_cloud_client().sdk_request(
          tenant= self.get_cloud_client().get_tenant_id(tenant= account, is_account= True), 
          lambda_sdk_command=lambda: resource_client.resources.list(filter="resourceType eq 'Microsoft.Compute/virtualMachineScaleSets'", expand="createdTime,changedTime,provisioningState")
        )
      }
    except:
      return []
        
  async def _process_account_data(self, account, loop, *args, **kwargs):
    client = ComputeManagementClient(credential= self.get_cloud_client().get_tenant_credential(tenant= self.get_cloud_client().get_tenant_id(tenant= account, is_account= True)), subscription_id= self.get_cloud_client().get_account_id(account= account))
    tasks = {
      "resource": loop.create_task(self.__process_get_resources_vmss(account= account))
    }

    await asyncio.wait(tasks.values())

    return {
      "account": account,
      "data": [ self.get_common().helper_type().dictionary().merge_dictionary([
          {},
          await self.get_base_return_data(
            account= self.get_cloud_client().serialize_resource(resource= account),
            resource_id= self.get_cloud_client().get_resource_id(resource= item),
            resource= item,
            region= self.get_cloud_client().get_resource_location(resource= item),
            resource_groups= [self.get_cloud_client().get_resource_group_from_resource(resource= item)],
          ),
          {
            "extra_resource": self.get_cloud_client().serialize_resource(tasks["resource"].result().get(self.get_cloud_client().get_resource_id(resource= item))),
            "extra_vmss_vms": [ self.get_cloud_client().serialize_resource(vm) for vm in client.virtual_machine_scale_set_vms.list(resource_group_name= self.get_cloud_client().get_resource_group_from_resource(item), virtual_machine_scale_set_name= item.name) ]
          },
          ]) for item in self.get_cloud_client().sdk_request(
          tenant= self.get_cloud_client().get_tenant_id(tenant= account, is_account= True), 
          lambda_sdk_command=lambda: client.virtual_machine_scale_sets.list_all()
        )
      ]
    }


"""The AWS dns Action. This will pull the AWS route53"""
from threemystic_cloud_data_client.cloud_providers.aws.client.actions.base_class.base import cloud_data_client_aws_client_action_base as base
import asyncio

class cloud_data_client_aws_client_action(base):
  def __init__(self, *args, **kwargs):
    super().__init__(
      data_action="dns",
      logger_name= "cloud_data_client_aws_client_action_dns",
      *args, **kwargs)
    
    
    self.data_id_name = "CertificateArn"
    
    self.arn_lambda = (lambda item: item["raw_item"][self.data_id_name])
    
    self.auto_region_resourcebytype= None # ["AWS::ACM::Certificate"]
    self.resource_group_filter = [
    {
      'Name': 'resource-type',
      'Values': [
        'AWS::ACM::Certificate',
      ]
    }
  ]
    
  
  async def __acm_certificates_tags(self, client, account, region, certificate_arn, *args, **kwargs):
      
      resource_list = self.get_cloud_client().general_boto_call_array(
          boto_call=lambda item: client.list_tags_for_certificate(**item),
          boto_params={"CertificateArn": certificate_arn},
          boto_nextkey = None,
          boto_key="Tags"
      )

      return resource_list
  
  async def __acm_certificates(self, client, *args, **kwargs):

    return self.get_cloud_client().general_boto_call_array(
      boto_call=lambda item: client.list_certificates(**item),
      boto_params={},
      boto_nextkey = "NextToken",
      boto_key="CertificateSummaryList"
    )
    
  
  async def _process_account_data_region(self, account, region, resource_groups, loop, *args, **kwargs):
    client = self.get_cloud_client().get_boto_client(client= 'acm',  account=account, region=region)

    tasks = {
      "main": loop.create_task(self.__acm_certificates(client= client)),
    }
    
    if len(tasks) > 0:
      await asyncio.wait(tasks.values())

    return {
      "region": region,
      "resource_groups": resource_groups,
      "data": [
        self.get_common().helper_type().dictionary().merge_dictionary([
          {},
          {
            "extra_tags": await self.__acm_certificates_tags(client= client, account= account, region= region, certificate_arn= item[self.data_id_name])
          },
          item
        ]) for item in tasks["main"].result()
        ]
    }



"""The AWS dns Action. This will pull the AWS route53"""
from threemystic_cloud_data_client.cloud_providers.aws.client.actions.base_class.base import cloud_data_client_aws_client_action_base as base
import asyncio

class cloud_data_client_aws_client_action(base):
  def __init__(self, *args, **kwargs):
    super().__init__(
      data_action="function",
      logger_name= "cloud_data_client_aws_client_action_function",
      *args, **kwargs)
    
    
    self.data_id_name = "Id"
    
    self.arn_lambda = (lambda item: self.get_cloud_client().get_resource_general_arn(
      resource_type= "route53",
      resource_type_sub= "hostedzone", **item # {region, account_id, resource_id}
    ))
    
    self.auto_region_resourcebytype= ["Amazon MemoryDB"]
    self.resource_group_filter = [
    {
      'Name': 'resource-type',
      'Values': [
        'AWS::Route53::HostedZone',
      ]
    }
  ]
    
  
  async def __hostedzones(self, client, *args, **kwargs):

    return self.get_cloud_client().general_boto_call_array(
      boto_call=lambda item: client.list_hosted_zones(**item),
      boto_params={},
      boto_nextkey = "NextMarker",
      boto_nextkey_param = "Marker",
      boto_key="HostedZones"
    )
    
  
  async def _process_account_data_region(self, account, region, resource_groups, loop, *args, **kwargs):
    client = self.get_cloud_client().get_boto_client(client= 'route53',  account=account, region=region)

    tasks = {
      "main": loop.create_task(self.__hostedzones(client= client)),
    }
    
    if len(tasks) > 0:
      await asyncio.wait(tasks.values())

    return {
      "region": region,
      "resource_groups": resource_groups,
      "data": [
        self.get_common().helper_type().dictionary().merge_dictionary([
          {},
          {
            "extra_tags": {}
          }, 
          item
        ]) for item in tasks["main"].result()
        ]
    }



from threemystic_cloud_data_client.cloud_providers.azure.client.actions.base_class.base import cloud_data_client_azure_client_action_base as base
import asyncio
from azure.mgmt.compute import ComputeManagementClient
from azure.mgmt.network import NetworkManagementClient
from azure.mgmt.resource import ResourceManagementClient
from azure.mgmt.network.models import LoadBalancer


class cloud_data_client_azure_client_action(base):
  def __init__(self, *args, **kwargs):
    super().__init__(
      data_action="vm", 
      logger_name= "cloud_data_client_azure_client_action_vm",
      *args, **kwargs)
  
  

  async def __process_get_resources_vm_public_ips(self, account, *args, **kwargs):    
    try:
        client = NetworkManagementClient(credential= self.get_cloud_client().get_tenant_credential(tenant= self.get_cloud_client().get_tenant_id(tenant= account, is_account= True)), subscription_id= self.get_cloud_client().get_account_id(account= account))

        return { self.get_cloud_client().get_resource_id(resource= public_ip):public_ip for public_ip in self.get_cloud_client().sdk_request(
           tenant= self.get_cloud_client().get_tenant_id(tenant= account, is_account= True), 
           lambda_sdk_command=lambda: client.public_ip_addresses.list_all()
          )
        }
           
    except Exception as err:
      return {} 

  async def __process_get_resources_vm_load_balancers(self, account, public_ips, *args, **kwargs):    
    try:
      client = NetworkManagementClient(credential= self.get_cloud_client().get_tenant_credential(tenant= self.get_cloud_client().get_tenant_id(tenant= account, is_account= True)), subscription_id= self.get_cloud_client().get_account_id(account= account))
      return_data = {}
      load_balancers = {self.get_cloud_client().get_resource_id(resource= lb): {"serialized": self.get_cloud_client().serialize_resource(resource= lb), "raw": lb} for lb in self.get_cloud_client().sdk_request(
          tenant= self.get_cloud_client().get_tenant_id(tenant= account, is_account= True), 
          lambda_sdk_command=lambda: client.load_balancers.list_all()
        )}

      if len(load_balancers) < 1:
        return {}
      

      load_balancers_front_end = {}

      for id_lb, lb in load_balancers.items():
        if lb["raw"].frontend_ip_configurations is None:
          continue
        for frontend_ip in lb["raw"].frontend_ip_configurations:
          if frontend_ip.load_balancing_rules is None:
            continue
          for frontend_ip_load_balancing_rule in frontend_ip.load_balancing_rules:
            if frontend_ip_load_balancing_rule is None:
              continue
            load_balancers_front_end[self.get_cloud_client().get_resource_id(resource= frontend_ip_load_balancing_rule)] = self.get_cloud_client().get_resource_id(resource= frontend_ip.public_ip_address)
      

      for id_lb, lb in load_balancers.items():
        if lb["raw"].frontend_ip_configurations is None:
          continue
        vm_data = {
          "load_balancer": lb["serialized"],
          "extra_public_ips": []
        }
        for backend_address_pool in lb["raw"].backend_address_pools:
          if backend_address_pool.load_balancing_rules is None:
            continue
            
          if backend_address_pool.load_balancer_backend_addresses is None:
            continue
          
          backend_frontend_connector = []
          for backend_address_load_balancing_rule in backend_address_pool.load_balancing_rules:            
            if backend_address_load_balancing_rule is None:
              continue

            if load_balancers_front_end.get(self.get_cloud_client().get_resource_id(resource= backend_address_load_balancing_rule)) is None:
              continue
            backend_frontend_connector.append(load_balancers_front_end.get(self.get_cloud_client().get_resource_id(resource= backend_address_load_balancing_rule)))
            
          for backend_address in backend_address_pool.load_balancer_backend_addresses:
            if backend_address.network_interface_ip_configuration is not None:
              if self.get_cloud_client().get_resource_id(resource= backend_address.network_interface_ip_configuration) is not None:
                key = self.get_cloud_client().get_resource_id(resource= backend_address.network_interface_ip_configuration)
                key = key[0:key.rfind("/ipconfigurations/")]
              else:
                key = backend_address.ip_address

              if self.get_common().helper_type().string().is_null_or_whitespace(string_value= key):
                continue

              return_data[key] = [
                {
                  "load_balancer": lb["serialized"],
                  "extra_public_ips": [
                    self.get_cloud_client().serialize_resource(resource= public_ips[ip_id])
                    for ip_id in backend_frontend_connector if public_ips.get(ip_id) is not None
                  ]
                }
              ]

          
      return return_data      
           
    except Exception as err:
      print(err)
      return {}

  async def __process_get_resources_vm_nics(self, account, public_ips, *args, **kwargs):    
    try:
      client = NetworkManagementClient(credential= self.get_cloud_client().get_tenant_credential(tenant= self.get_cloud_client().get_tenant_id(tenant= account, is_account= True)), subscription_id= self.get_cloud_client().get_account_id(account= account))
      return_data = {}
      for nic in self.get_cloud_client().sdk_request(
          tenant= self.get_cloud_client().get_tenant_id(tenant= account, is_account= True),
          lambda_sdk_command=lambda: client.network_interfaces.list_all()
        ):
          if nic.virtual_machine is None:
            continue
          
          nic_data = self.get_common().helper_type().dictionary().merge_dictionary([
            {},
            {
              "extra_public_ips": [
                self.get_cloud_client().serialize_resource(resource= public_ips[self.get_cloud_client().get_resource_id(resource= ip_config.public_ip_address)])
                for ip_config in nic.ip_configurations if self.get_cloud_client().get_resource_id(resource= ip_config.public_ip_address) is not None]
            },
            self.get_cloud_client().serialize_resource(resource= nic)
          ])
          if return_data.get( self.get_cloud_client().get_resource_id(resource= nic.virtual_machine)) is not None:
            return_data.get( self.get_cloud_client().get_resource_id(resource= nic.virtual_machine)).append(
              nic_data
            )
            continue
          

          return_data[self.get_cloud_client().get_resource_id(resource= nic.virtual_machine)] = [
            nic_data
          ]
      
      return return_data
          
        
           
    except Exception as err:
      return {}

  async def __process_get_resources_vm_availability_sets(self, client:ComputeManagementClient, account, *args, **kwargs):    
    try:
      return_data = {}

      for availability_set in  self.get_cloud_client().sdk_request(
          tenant= self.get_cloud_client().get_tenant_id(tenant= account, is_account= True), 
          lambda_sdk_command=lambda: client.availability_sets.list_by_subscription()
        ):
          for vm in availability_set.virtual_machines:
            return_data[self.get_cloud_client().get_resource_id(resource= vm)] = self.get_cloud_client().serialize_resource(resource= availability_set)
      
      return return_data
    except Exception as err:
      return {}
    
  async def __process_get_resources_vm(self, account, *args, **kwargs):
    resource_client = ResourceManagementClient(credential= self.get_cloud_client().get_tenant_credential(tenant= self.get_cloud_client().get_tenant_id(tenant= account, is_account= True)), subscription_id= self.get_cloud_client().get_account_id(account= account))
    try:
        return {self.get_cloud_client().get_resource_id(resource= resource): resource for resource in self.get_cloud_client().sdk_request(
           tenant= self.get_cloud_client().get_tenant_id(tenant= account, is_account= True), 
           lambda_sdk_command=lambda: resource_client.resources.list(filter="resourceType eq 'Microsoft.Compute/virtualMachines'", expand="createdTime,changedTime,provisioningState")
          )
        }
    except:
        return {}
    
  async def __process_get_resources_vm_statuses(self, account, client, *args, **kwargs):
    try:
        return {self.get_cloud_client().get_resource_id(resource= resource): self.get_cloud_client().serialize_resource(resource= resource) for resource in self.get_cloud_client().sdk_request(
           tenant= self.get_cloud_client().get_tenant_id(tenant= account, is_account= True),
           lambda_sdk_command=lambda: client.virtual_machines.list_all(
            status_only= True
           )
          )
        }
    except:
        return {}

  async def _process_account_data_get_vm_load_balancers(self, vm_nics, load_balancers_by_nics, *args, **kwargs):
    return_load_balancers = []
    
    for nic in vm_nics:
      if load_balancers_by_nics.get(self.get_cloud_client().get_resource_id(resource= nic)) is None:
        continue
        
      return_load_balancers += (load_balancers_by_nics.get(self.get_cloud_client().get_resource_id(resource= nic)))
      # I might look into the whole private IP agress later, the issue is I have to validate networks
      
    
    return return_load_balancers
  
  def __process_get_resources_vm_get_state(self, instanceView, *args, **kwargs):
    if instanceView is None:
      return None
    
    statuses = instanceView.get("statuses")
    if statuses is None:
      return None
    
    for status in statuses:
      if "powerstate/" in self.get_common().helper_type().string().set_case(string_value= status.get("code"), case= "lower"):
        return self.get_common().helper_type().string().set_case(string_value= self.get_common().helper_type().string().split(
          string_value= status.get("code"),
          separator= "[/]"
        )[-1], case= "upper")
      
    return None
  async def _process_account_data(self, account, loop, *args, **kwargs):
    client = ComputeManagementClient(credential= self.get_cloud_client().get_tenant_credential(tenant= self.get_cloud_client().get_tenant_id(tenant= account, is_account= True)), subscription_id= self.get_cloud_client().get_account_id(account= account))
    
    public_ips = await self.__process_get_resources_vm_public_ips(account= account)
    tasks = {
        "resource": loop.create_task(self.__process_get_resources_vm(account= account)),
        "vm_statuses": loop.create_task(self.__process_get_resources_vm_statuses(account= account, client= client)),
        "availability_sets": loop.create_task(self.__process_get_resources_vm_availability_sets(client= client, account= account)),
        "nics": loop.create_task(self.__process_get_resources_vm_nics(account= account, public_ips= public_ips)),
        "load_balancers": loop.create_task(self.__process_get_resources_vm_load_balancers(account= account, public_ips= public_ips)),
    }

    await asyncio.wait(tasks.values())

    return {
        "account": account,
        "data": [ self.get_common().helper_type().dictionary().merge_dictionary([
            {},
            await self.get_base_return_data(
              account= self.get_cloud_client().serialize_resource(resource= account),
              resource_id= self.get_cloud_client().get_resource_id(resource= item),
              resource= item,
              region= self.get_cloud_client().get_resource_location(resource= item),
              resource_groups= [self.get_cloud_client().get_resource_group_from_resource(resource= item)],
            ),     
            {
              "properties": {
                "instanceView": self.get_common().helper_type().general().get_container_value(
                  container= tasks["vm_statuses"].result().get(self.get_cloud_client().get_resource_id(resource= item)), 
                  value_key= ["properties", "instanceView"])
              }
            },
            {
              "extra_longlived": True,
              "extra_state": self.__process_get_resources_vm_get_state(instanceView= self.get_common().helper_type().general().get_container_value(
                  container= tasks["vm_statuses"].result().get(self.get_cloud_client().get_resource_id(resource= item)), 
                  value_key= ["properties", "instanceView"])),              
              "extra_resource": self.get_cloud_client().serialize_resource(tasks["resource"].result().get(self.get_cloud_client().get_resource_id(resource= item))),
              "extra_availability_set": tasks["availability_sets"].result().get(self.get_cloud_client().get_resource_id(resource= item)),
              "extra_nics": tasks["nics"].result().get(self.get_cloud_client().get_resource_id(resource= item)),
              "extra_load_balancers": await self._process_account_data_get_vm_load_balancers(
                vm_nics= tasks["nics"].result().get(self.get_cloud_client().get_resource_id(resource= item)),
                load_balancers_by_nics = tasks["load_balancers"].result()
              ),
            },
          ]) for item in self.get_cloud_client().sdk_request(
           tenant= self.get_cloud_client().get_tenant_id(tenant= account, is_account= True),
           lambda_sdk_command=lambda: client.virtual_machines.list_all()
          )
        ]
    }


from threemystic_common.base_class.base_provider import base
from threemystic_cloud_data_client.cli import cloud_data_client_cli
from threemystic_common.base_class.generate_data.generate_data_handlers import generate_data_handlers
from threemystic_cloud_client.cloud_providers.aws import cloud_client_aws as aws_client
from threemystic_cloud_client.cloud_providers.azure import cloud_client_azure as azure_client

class cloud_data_client_provider_base(base):
  def __init__(self, *args, **kwargs):
    if "provider" not in kwargs:
      kwargs["provider"] = self.get_default_provider()
    super().__init__(*args, **kwargs)
    
  def _setup_another_config(self, force_config = False, *args, **kwargs):
    if not force_config:
      response = self.get_common().generate_data().generate(
        generate_data_config = {
          "repeat_config": {
              "validation": lambda item: self.get_common().helper_type().bool().is_bool(check_value= item),
              "messages":{
                "validation": f"Valid options for Yes are: {self.get_common().helper_type().bool().is_true_values()}",
              },
              "conversion": lambda item: self.get_common().helper_type().bool().is_true(check_value= item),
              "desc": f"Do you want to setup another provider?: {self.get_common().helper_type().bool().is_true_values()}",
              "default": None,
              "handler": generate_data_handlers.get_handler(handler= "base"),
              "optional": True
          }
        }
      )

      if response is None:
        return
      
      if response.get("repeat_config") is None:
        return
      
      if response.get("repeat_config").get("formated") is not True:
        return
    
    print()
    print()
    print("-------------------------------------------------------------------------")
    print()
    print()

    cloud_data_client_cli().process_client_action("config")

  def update_general_config_completed(self, status, *args, **kwargs):
    self.get_config()["_config_process"] = status
    self._save_config()
  
  def is_cloud_client_config_completed(self, *args, **kwargs):
    cloud_client = None
    if self.get_provider() == "azure":
      cloud_client = azure_client( common= self.get_common(), logger= self.get_common().get_logger())
    if self.get_provider() == "aws":
      cloud_client = aws_client( common= self.get_common(), logger= self.get_common().get_logger())
    
    if cloud_client is not None:
      return cloud_client.is_provider_config_completed()

    return True
  
  def ensure_cloud_client_config_completed(self, *args, **kwargs):
    if self.is_cloud_client_config_completed():
      return True
    cloud_client = None
    if self.get_provider() == "azure":
      cloud_client = azure_client( common= self.get_common(), logger= self.get_common().get_logger())
    if self.get_provider() == "aws":
      cloud_client = aws_client( common= self.get_common(), logger= self.get_common().get_logger())
    
    if cloud_client is not None:
      cloud_client._setup_another_config()
    
    return True

    
  
  def is_general_config_completed(self, *args, **kwargs):    
    return self.get_config().get("_config_process") is True and self.is_cloud_client_config_completed()
  
  def get_main_directory_name(self, *args, **kwargs):
    return "data_client"  

  def __load_config(self, *args, **kwargs):
    config_data = self.get_common().helper_config().load(
      path= self.config_path(),
      config_type= "yaml"
    )
    if config_data is not None:
      return config_data
    
    return {}

  def reset_config_environment_data(self, *args, **kwargs):    
    self.get_config()["environment"] = {}
    self._save_config_environment_data()

  def has_tag_data_config(self, refresh = False, *args, **kwargs):
    
    if len(self.get_config_environment_data(refresh= refresh)) < 1:
      return False

    return True

  def get_config_environment_data(self, refresh = False, *args, **kwargs):
    if self.get_config(refresh= refresh).get("environment") is not None:
      return self.get_config().get("environment")
    
    self.get_config()["environment"] = {}
    self._save_config_environment_data()
     
    return self.get_config_environment_data(*args, **kwargs)

  def _update_config_environment_data(self,config_key, config_value, refresh = False,  *args, **kwargs):
     self.get_config_environment_data(refresh = refresh)[config_key] = config_value
     
  def _save_config_environment_data(self, *args, **kwargs):
     self._save_config()
  
  def get_environment_data_config_value(self, config_key, default_if_none = None, refresh = False, *args, **kwargs):
    config_value = self.get_config_environment_data(refresh= refresh).get(config_key)
    if config_value is not None:
      return config_value
    
    return default_if_none
  
  def config_path(self, *args, **kwargs):
    return self.get_common().get_threemystic_directory_config().joinpath(f"{self.get_main_directory_name()}/config")
  
  def get_config(self, refresh = False, *args, **kwargs):
    if hasattr(self, "_config_data") and not refresh:
      return self._config_data
    
    self._config_data = self.__load_config()    
    return self.get_config(*args, **kwargs)

  def _update_config(self,config_key, config_value, refresh= False,  *args, **kwargs):
     self.get_config(refresh = refresh)[config_key] = config_value
     
  def _save_config(self, *args, **kwargs):
     if not self.config_path().parent.exists():
       self.config_path().parent.mkdir(parents= True)
     self.config_path().write_text(
      data= self.get_common().helper_yaml().dumps(data= self.get_config())
     )
     self.get_config(refresh = True) 
  
  def get_config_value(self, config_key, default_if_none = None, refresh = False, *args, **kwargs):
    config_value = self.get_config(refresh= refresh).get(config_key)
    if config_value is not None:
      return config_value
    
    return default_if_none
  
  def set_default_fiscal_year_start(self, value, refresh = False, *args, **kwargs):
    if self.get_common().helper_type().string().is_null_or_whitespace(string_value=value):
      value = self.get_default_fiscal_year_start(refresh= refresh)

    self.get_config(refresh= refresh)["fiscal_year_start"] = value
    self._save_config()
  
  def get_default_currency(self, refresh = False, *args, **kwargs):
    if self.get_common().helper_type().string().is_null_or_whitespace(string_value=self.get_config(refresh= refresh).get("currency")):
      return "usd"
    
    return self.get_config(refresh= refresh).get("currency")
    
  def get_default_fiscal_year_start(self, refresh = False, *args, **kwargs):
    if self.get_common().helper_type().string().is_null_or_whitespace(string_value=self.get_config(refresh= refresh).get("fiscal_year_start")):
      return "01/01"
    
    if self.get_common().helper_type().datetime().datetime_from_string(dt_string= f'{self.get_common().helper_type().datetime().get().year}/{self.get_config().get("default_output_format")}') is not None:
      return "01/01"
    
    return self.get_common().helper_type().string().set_case(string_value= self.get_config().get("fiscal_year_start"), case= "lower")
  
  def get_default_output_format(self, refresh = False, *args, **kwargs):
    
    if self.get_common().helper_type().string().is_null_or_whitespace(string_value=self.get_config(refresh= refresh).get("default_output_format")):
      return "json"
    
    if self.get_common().helper_type().string().set_case(string_value= self.get_config().get("default_output_format"), case= "lower") not in self.get_supported_output_format():
      return "json"
    
    return self.get_common().helper_type().string().set_case(string_value= self.get_config().get("default_output_format"), case= "lower")
  
  def set_default_output_format(self, value, refresh = False, *args, **kwargs):
    if self.get_common().helper_type().string().is_null_or_whitespace(string_value=value):
      value = self.get_default_output_format(refresh= refresh)

    self.get_config(refresh= refresh)["default_output_format"] = value
    self._save_config()
  
  def get_default_provider(self, refresh = False, *args, **kwargs):
    return self.get_config(refresh= refresh).get("default_provider")
  
  def set_default_provider(self, value, refresh = False, *args, **kwargs):
    self.get_config(refresh= refresh)["default_provider"] = value
    self._save_config()

  def action_config(self, *args, **kwargs):
    print("Provider config not configured")

  def action_data(self, *args, **kwargs):
    print("Provider config not configured")
  
  
    

  
  




from threemystic_cloud_data_client.cloud_providers.general.config.base_class.base import cloud_data_client_general_config_base as base
from threemystic_common.base_class.generate_data.generate_data_handlers import generate_data_handlers
from threemystic_cloud_data_client.cloud_providers.general.config.step_2 import cloud_data_client_general_config_step_2 as step



class cloud_data_client_general_config_step_1(base):
  def __init__(self, *args, **kwargs):
    super().__init__(logger_name= "cloud_data_client_general_config_step_1", *args, **kwargs)
    

  def step(self, *args, **kwargs):
    if not super().step(run_base_config= True):
      return
    
    response = self.get_common().generate_data().generate(
      generate_data_config = {
        "default_provider": {
            "validation": lambda item: self.get_common().helper_type().string().set_case(string_value= item, case= "lower") in self.get_supported_providers(),
            "messages":{
              "validation": f"Valid options are: {self.get_supported_providers()}",
            },
            "conversion": lambda item: self.get_common().helper_type().string().set_case(string_value= item, case= "lower"),
            "desc": f"What do you want as the the default provider? \nValid Options: {self.get_supported_providers()}",
            "default": self.get_default_provider(),
            "handler": generate_data_handlers.get_handler(handler= "base"),
            "optional": True
        },        
        "default_output_format": {
            "validation": lambda item: self.get_common().helper_type().string().set_case(string_value= item, case= "lower") in self.get_supported_output_format(),
            "messages":{
              "validation": f"Valid options are: {self.get_supported_output_format()}",
            },
            "conversion": lambda item: self.get_common().helper_type().string().set_case(string_value= item, case= "lower"),
            "desc": f"What is the default output format.\nValid Options: {self.get_supported_output_format()}",
            "default": self.get_default_output_format(),
            "handler": generate_data_handlers.get_handler(handler= "base"),
            "optional": not self.get_common().helper_type().string().is_null_or_whitespace(string_value= self.get_default_output_format())
        },        
        "fiscal_year_start": {
            "validation": lambda item: not self.get_common().helper_type().string().is_null_or_whitespace(string_value= item) and self.get_common().helper_type().datetime().datetime_from_string(dt_string= f'{self.get_common().helper_type().datetime().get().year}/{item}') is not None,
            "messages":{
              "validation": f"Please provide a valid date in the following format mm/dd",
            },
            "conversion": lambda item: self.get_common().helper_type().string().set_case(string_value= item, case= "lower"),
            "desc": f"What is the fiscal year start date? ex mm/dd (10/01)",
            "default": self.get_default_fiscal_year_start(),
            "handler": generate_data_handlers.get_handler(handler= "base"),
            "optional": not self.get_common().helper_type().string().is_null_or_whitespace(string_value= self.get_default_fiscal_year_start())
        },        
        "currency": {
            "validation": lambda item: not self.get_common().helper_type().string().is_null_or_whitespace(string_value= item),
            "messages":{
              "validation": f"Please provide a valid date in the following format mm/dd",
            },
            "conversion": lambda item: self.get_common().helper_type().string().trim(string_value= self.get_common().helper_type().string().set_case(string_value= item, case= "lower")),
            "desc": f"What should currency be converted to?",
            "default": self.get_default_currency(),
            "handler": generate_data_handlers.get_handler(handler= "base"),
            "optional": True
        }
      }
    )

    if(response is not None):
      for key, item in response.items():
        self._update_config(config_key= key, config_value= item.get("formated"))
      self._save_config()
      
      print("-----------------------------")
      print()
      print()
      print("Base Configuration is updated")
      print()
      print()
      print("-----------------------------")

      next_step = step(common= self.get_common(), logger= self.get_logger())
      
      return next_step.step()
    

    
      print("-----------------------------")
      print()
      print()
      print("Base Configuration NOT updated")
      print()
      print()
      print("-----------------------------")
    


    
    
  



"""The AWS vm Action. This will pull the AWS EC2s"""
from threemystic_cloud_data_client.cloud_providers.aws.client.actions.base_class.base import cloud_data_client_aws_client_action_base as base
import asyncio

class cloud_data_client_aws_client_action(base):
  def __init__(self, *args, **kwargs):
    super().__init__(
      data_action="database",
      logger_name= "cloud_data_client_aws_client_action_database",
      *args, **kwargs)
    
    
    self.data_id_name = "DBInstanceIdentifier"
    
    self.arn_lambda = (lambda item: self.get_cloud_client().get_resource_general_arn(
      resource_type= "rds",
      resource_type_sub= "db", **item # {region, account_id, resource_id}
    ))
    
    self.auto_region_resourcebytype= ["Amazon Relational Database Service"]
    self.resource_group_filter = [
    {
      'Name': 'resource-type',
      'Values': [
        'AWS::RDS::DBCluster',
      ]
    },
    {
      'Name': 'resource-type',
      'Values': [
        'AWS::RDS::DBInstance',
      ]
    }
  ]
    
  
  async def __clusters(self, client, account, region, resource_groups, *args, **kwargs):

    db_clusters = self.get_cloud_client().general_boto_call_array(
      boto_call=lambda item: client.describe_db_clusters(**item),
      boto_params={},
      boto_nextkey = "Marker",
      boto_key="DBClusters"
    )

    cluster_search = {}
    db_clusters_byid = {}
    for cluster in db_clusters:
      resource_arn = (self.get_cloud_client().get_resource_general_arn(
        resource_type= "rds",
        resource_type_sub= "cluster", 
        region= region,
        account_id= self.get_cloud_client().get_account_id(account= account),
        resource_id= cluster.get("DBClusterIdentifier")
      ))
      
      db_clusters_byid[cluster["DBClusterIdentifier"]] =  await self.get_base_return_data(
        account= account,
        resource_id= cluster.get("DBClusterIdentifier"),
        resource= cluster,
        region= region,
        resource_groups= resource_groups.get(resource_arn),
      )
      
      if cluster.get("DBClusterMembers") is None:
        continue

      if len(cluster.get("DBClusterMembers")) < 1:
        continue

      for member in cluster.get("DBClusterMembers"):
        cluster_search[member["DBInstanceIdentifier"]] = cluster["DBClusterIdentifier"]
    
    return {
      "clusters": db_clusters_byid,
      "cluster_dbmembers": cluster_search
    }
  
  async def __get_databases(self, client, *args, **kwargs):
    return self.get_cloud_client().general_boto_call_array(
      boto_call=lambda item: client.describe_db_instances(**item),
      boto_params={},
      boto_nextkey = "Marker",
      boto_key="DBInstances"
    ) 
  
  def __get_maintenance_key(self, ResourceIdentifier, *args, **kwargs):
    resource_identifier_split = ResourceIdentifier.split(":")

    return f"{resource_identifier_split[-2]}:{resource_identifier_split[-1]}"
  
  def __get_cluster_data_db(self, clusters, db, *args, **kwargs):
    if not db["DBInstanceIdentifier"] in clusters["cluster_dbmembers"]:
      return None
    
    return clusters["clusters"][clusters["cluster_dbmembers"][db["DBInstanceIdentifier"]]]
  
  async def __get_maintenance_actions(self, client, *args, **kwargs):
    pending_maintenance_actions = self.get_cloud_client().general_boto_call_array(
      boto_call=lambda item: client.describe_pending_maintenance_actions(**item),
      boto_params={},
      boto_nextkey = "Marker",
      boto_key="PendingMaintenanceActions"
    )

    pending_maintenance_actions_db = {}
    for action in pending_maintenance_actions:
      action_key = self.__get_maintenance_key(action["ResourceIdentifier"])
      if pending_maintenance_actions_db.get(action_key) is None:
        pending_maintenance_actions_db[action_key] = []

      pending_maintenance_actions_db[action_key].append(action)
    
    return pending_maintenance_actions_db
  
  async def _process_account_data_region(self, account, region, resource_groups, loop, *args, **kwargs):
    client = self.get_cloud_client().get_boto_client(client= 'rds',  account=account, region=region)

    tasks = {
      "clusters": loop.create_task(self.__clusters(client= client, account= account, region= region, resource_groups= resource_groups)),
      "databases": loop.create_task(self.__get_databases(client= client)),
      "pending_maintenance_actions": loop.create_task(self.__get_maintenance_actions(client= client)),
    }
    
    if len(tasks) > 0:
      await asyncio.wait(tasks.values())

    return {
      "region": region,
      "resource_groups": resource_groups,
      "data": [
        self.get_common().helper_type().dictionary().merge_dictionary([
          {},
          {
          "extra_cluster": self.__get_cluster_data_db(clusters= tasks["clusters"].result(), db= item),
          "extra_pending_maintenance_actions": tasks["pending_maintenance_actions"].result().get(item["DBInstanceIdentifier"]),
          },
          item
        ]) for item in tasks["databases"].result()
      ]
    }



from threemystic_cloud_data_client.cloud_providers.base_class.base_data import cloud_data_client_provider_base_data as base

class cloud_data_client_azure_client_action_base(base):
  def __init__(self, *args, **kwargs):
    super().__init__(provider= "azure", *args, **kwargs)  

  def get_accounts(self, *args, **kwargs):

    if len(self.get_runparam_key(data_key= "data_accounts", default_value= [])) < 1:
      return [ 
        account for account in self.get_cloud_client().get_accounts() 
        if account.resource_container and account.subscription_id == "5c49443f-ce4b-470d-8ba8-c9571c1de06d" ]
    
    return [ 
        account for account in self.get_cloud_client().get_accounts() 
        if( account.resource_container and 
            self.get_cloud_client().get_account_id(account= account) in self.get_runparam_key(data_key= "data_accounts", default_value= []) and 
            not f'-{self.get_cloud_client().get_account_id(account= account)}' in self.get_runparam_key(data_key= "data_accounts", default_value= [])
          )
        ]


from threemystic_cloud_data_client.cloud_providers.aws.base_class.base import cloud_data_client_provider_aws_base as base


class cloud_data_client_aws(base):
  def __init__(self, *args, **kwargs):
    super().__init__(logger_name= "cloud_data_client_aws", *args, **kwargs)

  def action_config(self, *args, **kwargs): 
    
    from threemystic_cloud_data_client.cloud_providers.aws.config.step_1 import cloud_data_client_aws_config_step_1 as step
    next_step = step(common= self.get_common(), logger= self.get_logger())
    
    next_step.step()


  
    
    
  



"""The AWS vm Action. This will pull the AWS EC2s"""
from threemystic_cloud_data_client.cloud_providers.aws.client.actions.base_class.base import cloud_data_client_aws_client_action_base as base
import asyncio
from threemystic_cloud_data_client.cloud_providers.aws.client.actions.vmimage import cloud_data_client_aws_client_action as vmimage_data

class cloud_data_client_aws_client_action(base):
  def __init__(self, *args, **kwargs):
    super().__init__(
      data_action="vm",
      logger_name= "cloud_data_client_aws_client_action_vm",
      *args, **kwargs)
    
    
    self.data_id_name = "InstanceId"
    
    self.arn_lambda = (lambda item: self.get_cloud_client().get_resource_general_arn(
      resource_type= "ec2",
      resource_type_sub= "instance", **item # {region, account_id, resource_id}
    ))
    
    self.auto_region_resourcebytype= ["Amazon Elastic Compute Cloud - Compute"]
    self.resource_group_filter = [
    {
      'Name': 'resource-type',
      'Values': [
        'AWS::EC2::Instance',
      ]
    }
  ]
    
  
  async def _pre_load_main_process(self, pool, *args, **kwarg):
    self.required_extra_data["instances_information"] = {}
    if self.required_extra_data.get("ami") is None:
      base_action_params = {
        "cloud_data_client": self._get_cloud_data_client_raw(),
        "common": self.get_common(),
        "logger": self.get_common().get_logger()
      }
      
      ami_data = vmimage_data(
        **base_action_params
      )
      tmp_ami_data = await ami_data.main(pool=pool, run_params= self.get_runparam_key(data_key= None))
      self.required_extra_data["ami"] = await ami_data.get_data_by_id(results= tmp_ami_data, id_override = "extra_id_only")

  def _process_preloaded_data(self, preloaded_data = None, *args, **kwargs):
    self.required_extra_data = {}
    if preloaded_data is not None:
      if preloaded_data.get("ami") is not None:        
        if len(preloaded_data.get("ami")) > 0:
          self.required_extra_data["ami"] = preloaded_data.get("ami")
    
  def __get_asg_extra_data(self, instance_tags = None, asg_data = None, *args, **kwargs):
    if asg_data is not None:
      return {"arn": asg_data["AutoScalingGroupARN"], "name":  asg_data["AutoScalingGroupName"]}
    
    if instance_tags is not None:
      if instance_tags.get("aws:autoscaling:groupName") is not None:
        return {"name":  instance_tags.get("aws:autoscaling:groupName")}

    return None
  
  async def __get_asg_by_instance(self, account, region, *args, **kwargs):

    auto_scaling_client = self.get_cloud_client().get_boto_client(
      client= "autoscaling",
      account= account,
      region= region
    )

    groups = self.get_cloud_client().general_boto_call_array(
      boto_call=lambda item: auto_scaling_client.describe_auto_scaling_groups(**item),
      boto_params={ },
      boto_nextkey = "NextToken",
      boto_key="AutoScalingGroups",
      logger = self.get_common().get_logger()
    )

    return_data = {}

    for asg in groups:
      for instance in asg["Instances"]:
        return_data[instance["InstanceId"]] = self.__get_asg_extra_data(asg_data= asg)
    
    return return_data
  
  async def __process_account_region_lb_instances_targethealth(self, alb_client, targetgroup_arn, *args, **kwargs):
    return self.get_cloud_client().general_boto_call_array(
      boto_call=lambda: alb_client.describe_target_health(TargetGroupArn=targetgroup_arn),
      boto_params= None,
      boto_nextkey = None,
      boto_key="TargetHealthDescriptions",
      logger = self.get_common().get_logger()
    )

  async def __process_account_region_lb_instances_targetgroups(self, alb_client, lb_arn, *args, **kwargs):
    return self.get_cloud_client().general_boto_call_array(
      boto_call=lambda: alb_client.describe_target_groups(LoadBalancerArn=lb_arn),
      boto_params= None,
      boto_nextkey = None,
      boto_key="TargetGroups",
      logger = self.get_common().get_logger()
    )
  
  async def __process_account_region_lb_instances(self, account, region, loop, *args, **kwargs):

      lb_instances = {}
      
      elbClassic = self.get_cloud_client().get_boto_client(
        client= "elb",
        account= account,
        region= region
      )

      elbs = self.get_cloud_client().general_boto_call_array(
        boto_call=lambda item: elbClassic.describe_load_balancers(),
        boto_params={},
        boto_nextkey = None,
        boto_key="LoadBalancerDescriptions",
        logger = self.get_common().get_logger()
      )
      
      for elb in elbs:
        num_instances = len(elb['Instances'])
        if num_instances < 1:
          continue
        for i in elb['Instances']:
          if not i['InstanceId'] in lb_instances:
            lb_instances[i['InstanceId']] = {}
          if elb['LoadBalancerName'] in lb_instances[i['InstanceId']]:
            continue
          lb_instances[i['InstanceId']][elb['LoadBalancerName']] = ({"instance": i['InstanceId'], "lbType": "ELB", "lbName": elb['LoadBalancerName'], "lbDNSName": elb['DNSName']})
      
      alb = self.get_cloud_client().get_boto_client(
        client= "elbv2",
        account= account,
        region= region
      )

      albs = self.get_cloud_client().general_boto_call_array(
        boto_call=lambda item: alb.describe_load_balancers(),
        boto_params={},
        boto_nextkey = None,
        boto_key="LoadBalancers",
        logger = self.get_common().get_logger()
      )
      if len(albs) < 1:
        return lb_instances

      targetGroups = { 
        lb['LoadBalancerArn']: loop.create_task(self.__process_account_region_lb_instances_targetgroups(alb_client= alb, lb_arn= lb['LoadBalancerArn'])) for lb in albs
      }
      
      if len(targetGroups) < 1:
        return lb_instances
      await asyncio.wait(targetGroups.values())
    
      for lb in albs:
        targets_health = {
          group['TargetGroupArn']: loop.create_task(self.__process_account_region_lb_instances_targethealth(alb_client= alb, targetgroup_arn= group['TargetGroupArn'])) for group in targetGroups[lb['LoadBalancerArn']].result()
        }  
        if len(targets_health) < 1:
         continue
        await asyncio.wait(targets_health.values())

        for _, targets in targets_health.items():
          for target in targets.result():
            if target.get('Target') is None or len(target['Target']) < 1:
              continue

            if not target['Target']['Id'] in lb_instances:
              lb_instances[target['Target']['Id']] = {}
            if lb['LoadBalancerName'] in lb_instances[target['Target']['Id']]:
              continue
            lb_instances[target['Target']['Id']][lb['LoadBalancerName']] = ({"instance": target['Target']['Id'], "lbType": "ALB", "lbName": lb['LoadBalancerName'], "lbDNSName": lb['DNSName']})        
      
      return lb_instances
  
  async def __get_ec2_volume_details(self, ec2_client, *args, **kwargs):
        
      volumne_data = self.get_cloud_client().general_boto_call_array(
        boto_call=lambda item: ec2_client.describe_volumes(**item),
        boto_params={},
        boto_nextkey = "NextToken",
        boto_key="Volumes"
      )

      return { data["VolumeId"]:data for data in volumne_data }
  
  async def __getImageInfo(self, image_id, *args, **kwargs):
    if image_id in self.required_extra_data["ami"]:
      return {
        "Id": image_id,
        "Name": self.required_extra_data["ami"][image_id]["Name"],
        "Description": self.required_extra_data["ami"].get("Description")
      }


    return {
      "Id": image_id,
    }
  
  async def __get_instance_attribute(self, client, instance, loop, attribute = None, *args, **kwargs):
    # Full attribute list
    # 'instanceType','kernel','ramdisk','userData','disableApiTermination','instanceInitiatedShutdownBehavior','rootDeviceName','blockDeviceMapping','productCodes','sourceDestCheck','groupSet','ebsOptimized','sriovNetSupport','enaSupport','enclaveOptions','disableApiStop'
    # to use groupSet you need to specify an interface ID when multiple interfaces are attached

    full_attribute_list = ['instanceType','kernel','ramdisk','userData','disableApiTermination','instanceInitiatedShutdownBehavior','rootDeviceName','blockDeviceMapping','productCodes','sourceDestCheck','groupSet','ebsOptimized','sriovNetSupport','enaSupport','enclaveOptions','disableApiStop']
    instance_attributes_dict = {attr.lower():attr for attr in full_attribute_list}
    if attribute is None:
      attribute = ['instanceInitiatedShutdownBehavior','disableApiTermination','disableApiStop']
      
    elif self.get_common().helper_type().general().is_type(obj=attribute, type_check=str):
      attribute = [ attr.strip() for attr in attribute.split(",") if not self.get_common().helper_type().string().is_null_or_whitespace(string_value= attribute) ]

    if len(attribute) < 1:
      return None
    
    if len(attribute) > 1:
      tasks = {
        attr.lower(): loop.create_task(self.__get_instance_attribute(client= client, instance= instance, loop= loop, attribute= attr, *args, **kwargs)) for attr in attribute

      }
      await asyncio.wait(tasks.values())
      return {attr:task.result() for attr,task in tasks.items()}


    attribute = attribute[0].lower()
    if attribute not in instance_attributes_dict:
      return None

    if attribute == "groupset" and instance.get("NetworkInterfaces") is not None and len(instance.get("NetworkInterfaces")) > 1:
        return {
          "extra_instance_attribute": instance_attributes_dict[attribute],
          "Value": "Multiple Interfaces"
        }
    try:        
      return self.get_common().helper_type().dictionary().merge_dictionary([
        {},
        {"extra_instance_attribute": instance_attributes_dict[attribute]},
        await self.get_cloud_client().async_general_boto_call_single(
          boto_call=lambda item: client.describe_instance_attribute(**item),
          boto_params={"Attribute": instance_attributes_dict[attribute], "InstanceId": instance["InstanceId"]},
          boto_nextkey = None,
          boto_key= lambda item: item.get(self.get_common().helper_type().list().find_item(data= item.keys(), filter= lambda search: search.lower() != "instanceid" and search.lower() != "responsemetadata" )),
          error_codes_raise= ["InvalidInstanceID.NotFound"],
          logger = self.get_common().get_logger()
        )
      ])
    except Exception as ex:
      raise self.get_common().exception().exception(
        exception_type = "generic"
      ).exception(
        logger= self.get_logger(),
        name = f"Check Instance: {instance['InstanceId']} - {kwargs['account']}",
        message = f"unable to get extra_instance_attribute {instance_attributes_dict[attribute]}",
        exception= ex
      )
  
  async def __getSSMInfo(self, ssm_client, account, region, instance, filter = [], *args, **kwargs):

    if self.get_cloud_client().get_account_id(account) not in self.required_extra_data["instances_information"]:
      instances_information = self.get_cloud_client().general_boto_call_array(
        boto_call=lambda item: ssm_client.describe_instance_information(**item),
        boto_params={"Filters": filter},
        boto_nextkey = "NextToken",
        boto_key="InstanceInformationList",
        logger = self.get_common().get_logger()
      )

      self.required_extra_data["instances_information"][self.get_cloud_client().get_account_id(account)] = {instance["InstanceId"]:instance for instance in instances_information}

    if instance["InstanceId"] not in self.required_extra_data["instances_information"][self.get_cloud_client().get_account_id(account)]:
      return {}

    instance_information = self.required_extra_data["instances_information"][self.get_cloud_client().get_account_id(account)][instance["InstanceId"]]

    return {
        "PlatformName": instance_information.get('PlatformName'),
        "PlatformVersion": instance_information.get('PlatformVersion'),
        "PingStatus": instance_information.get('PingStatus'),
        "PingTime": instance_information.get('LastPingDateTime'),
        "Version": instance_information.get('AgentVersion'),
    } 
  
  async def __get_instance_withextra_data(self, ec2_client, ssm_client, task_asg_by_instance, account, region, lb_instances, instance, loop, *args, **kwargs):
      
      
      extra_data_tasks = {
        "extra_image": loop.create_task(self.__getImageInfo(image_id= instance["ImageId"])),
        "extra_instance_attributes": loop.create_task(self.__get_instance_attribute(account=account, client= ec2_client, instance=instance, loop= loop)),
        "extra_ssminfo": loop.create_task(self.__getSSMInfo(account=account, region=region, ssm_client=ssm_client, instance=instance)),
        "_ignore_extra_asg": task_asg_by_instance
        
      }      

      extra_data = {
        "extra_platform": instance["Platform"] if not self.get_common().helper_type().string().is_null_or_whitespace(string_value= instance.get("Platform")) else "Linux",
        "extra_subnet": instance["SubnetId"] if instance.get("SubnetId") is not None else "EC2Classic",
        "extra_vpc": instance["VpcId"] if instance.get("VpcId") is not None else "EC2Classic",
        "extra_lb": lb_instances.get(instance["InstanceId"]) if lb_instances is not None else None,
      }

      await asyncio.wait(extra_data_tasks.values())

      for key,item in extra_data_tasks.items():
        if key.lower().startswith("_ignore_"):
          continue
        extra_data[key]= item.result()

      extra_data["extra_asg"] = extra_data_tasks["_ignore_extra_asg"].result().get(instance["InstanceId"]) if extra_data_tasks["_ignore_extra_asg"].result() is not None else None
      if extra_data["extra_asg"] is None:
        extra_data["extra_asg"] = self.__get_asg_extra_data(instance_tags= self.get_cloud_client().get_resource_tags_as_dictionary(resource= instance))
      
      return self.get_common().helper_type().dictionary().merge_dictionary([{}, extra_data, instance])
  
  def __is_ec2_ebs_encrypted_state(self, running_state, block_device_details):
    if block_device_details is None:
      if self.get_common().helper_type().string().is_null_or_whitespace(running_state):
          return None
      return running_state.lower()
    
    if running_state is None: 
      if not block_device_details["Encrypted"]:
        return "none"
      return "all"
    
    if running_state.lower() == "all"  and block_device_details["Encrypted"]:
      return "all"
    
    return "partial"
  
  async def __get_ec2_instances(self, ec2_client, ssm_client, task_asg_by_instance, account, region, loop, *args, **kwargs):
    instances_reservations = self.get_cloud_client().general_boto_call_array(
      boto_call=lambda item: ec2_client.describe_instances(**item),
      boto_params={},
      boto_nextkey = "NextToken",
      boto_key="Reservations",
      logger = self.get_common().get_logger()
    )

    if len(instances_reservations) < 1:
      return []

    pre_process_tasks = {
      "lb_instances": loop.create_task(self.__process_account_region_lb_instances(account, region, loop, *args, **kwargs)),
      "volume_details": loop.create_task(self.__get_ec2_volume_details(ec2_client= ec2_client, *args, **kwargs))

    }
    await asyncio.wait(pre_process_tasks.values())
    
  
    instance_tasks = []

    instance_ebs_encrypted = {}
    for reservation in instances_reservations:
      for instance in reservation["Instances"]:
        if (instance["State"].get("Name")).lower() == "terminated":
          continue
        
        instance_ebs_encrypted[instance[self.data_id_name]] = None
        instance_tasks.append(loop.create_task(self.__get_instance_withextra_data(ec2_client=ec2_client, ssm_client= ssm_client, task_asg_by_instance= task_asg_by_instance, account= account, region= region, lb_instances=pre_process_tasks["lb_instances"].result(), instance= instance, loop= loop)))
        if instance.get("BlockDeviceMappings") is not None:
          for block_device in instance["BlockDeviceMappings"]:
            if block_device.get("Ebs") is None:
              instance_ebs_encrypted[instance[self.data_id_name]] = self.__is_ec2_ebs_encrypted_state(
                running_state= instance_ebs_encrypted[instance[self.data_id_name]],
                block_device_details= None
              )
              continue
            
            block_device["Ebs"]["extra_details"] = pre_process_tasks["volume_details"].result().get(block_device["Ebs"]["VolumeId"])
            instance_ebs_encrypted[instance[self.data_id_name]] = self.__is_ec2_ebs_encrypted_state(
              running_state= instance_ebs_encrypted[instance[self.data_id_name]],
              block_device_details= block_device["Ebs"]["extra_details"]
            )
          
          instance["extra_ebs_encrypted_state"] = instance_ebs_encrypted[instance[self.data_id_name]]


    await asyncio.wait(instance_tasks)
    return [ instance.result() for instance in instance_tasks]
  
  async def _process_account_data_region(self, account, region, resource_groups, loop, *args, **kwargs):
    ec2_client = self.get_cloud_client().get_boto_client(
      client= "ec2",
      account= account,
      region= region
    )
    ssm_client = self.get_cloud_client().get_boto_client(
      client= "ssm",
      account= account,
      region= region
    )

    task_asg_by_instance = loop.create_task(self.__get_asg_by_instance(account, region))
    tasks = {
      "instances": loop.create_task(self.__get_ec2_instances(ec2_client= ec2_client, ssm_client= ssm_client, task_asg_by_instance= task_asg_by_instance, account= account, region= region, loop= loop, **kwargs))
    }
    
    if len(tasks) > 0:
      await asyncio.wait(tasks.values())

    return {
      "region": region,
      "resource_groups": resource_groups,
      "data": tasks["instances"].result()
    }



"""The AWS VMImage Action. This will pull the AWS AMIs"""
from threemystic_cloud_data_client.cloud_providers.aws.client.actions.base_class.base import cloud_data_client_aws_client_action_base as base
import asyncio


class cloud_data_client_aws_client_action(base):
  def __init__(self, *args, **kwargs):
    super().__init__(
      data_action="vmimage",
      logger_name= "cloud_data_client_aws_client_action_vmimage",
      *args, **kwargs)
    
    
    self.data_id_name = "ImageId"
    
    self.arn_lambda = (lambda item: self.get_cloud_client().get_resource_general_arn(
      resource_type= "ec2",
      resource_type_sub= "image", **item # {region, account_id, resource_id}
    ))
    
    self.auto_region_resourcebytype= ["Amazon Elastic Compute Cloud - Compute"]
    self.resource_group_filter = [
      {
        'Name': 'resource-type',
        'Values': [
          'AWS::EC2::Image',
        ]
      }
    ]
  
  
  
  async def _process_account_data_region(self, account, region, resource_groups, loop, *args, **kwargs):
    ec2_client = self.get_cloud_client().get_boto_client(
      client= "ec2",
      account= account,
      region= region
    )

    awsdata = self.get_cloud_client().general_boto_call_array(
      boto_call=lambda item: ec2_client.describe_images(**item),
      boto_params= {"Owners": ["self"], "IncludeDeprecated": True},
      boto_nextkey = "NextToken",
      boto_key="Images"
    )

    return {
      "region": region,
      "resource_groups": resource_groups,
      "data": awsdata
    }



"""The AWS database Action. This will pull the AWS rds"""
from threemystic_cloud_data_client.cloud_providers.aws.client.actions.base_class.base import cloud_data_client_aws_client_action_base as base
import asyncio

#add elasticache
class cloud_data_client_aws_client_action(base):
  def __init__(self, *args, **kwargs):
    super().__init__(
      data_action="memorydb",
      logger_name= "cloud_data_client_aws_client_action_memorydb",
      *args, **kwargs)
    
    
    self.data_id_name = "ARN"
    
    self.arn_lambda = (lambda item: self.get_cloud_client().get_resource_general_arn(
      resource_type= "memorydb",
      resource_type_sub= "cluster", **item # {region, account_id, resource_id}
    ))
    
    self.auto_region_resourcebytype= ["Amazon MemoryDB", "Amazon Elastic Compute Cloud - Compute", "Amazon Relational Database Service"]
    self.resource_group_filter = [
    {
      'Name': 'resource-type',
      'Values': [
        'AWS::MemoryDB::Cluster',
        'AWS::ElastiCache::CacheCluster'
      ]
    }
  ]
    
  
  async def __get_memorydb_tags(self, client, account, region, *args, **kwargs):
      
      resource_list = self.get_cloud_client().general_boto_call_array(
          boto_call=lambda item: client.get_resources(**item),
          boto_params={"ResourceTypeFilters": ['memorydb']},
          boto_nextkey = "PaginationToken",
          boto_key="ResourceTagMappingList"
      )

      return {
        resource["ResourceARN"].lower():resource["Tags"] for resource in resource_list
      }

  async def __get_clusters(self, client, *args, **kwargs):
    return self.get_cloud_client().general_boto_call_array(
      boto_call=lambda item: client.describe_clusters(**item),
      boto_params={"ShowShardDetails": True},
      boto_nextkey = "NextToken",
      boto_key="Clusters"
    )
  
  async def __get_elasticache_tags(self, client, account, region, cache, *args, **kwargs):
      
    try:
      return await self.get_cloud_client().async_general_boto_call_array(
          boto_call=lambda item: client.list_tags_for_resource(**item),
          boto_params={"ResourceName": cache[self.data_id_name]},
          boto_nextkey = "Marker",
          boto_key="TagList"
      )
    except Exception as err:
      self.get_common().get_logger().exception(msg= f"__get_elasticache_tags: {err}", extra= {"exception": err})
      return []
  
  async def __get_elasticache(self, client, *args, **kwargs):
    return self.get_cloud_client().general_boto_call_array(
      boto_call=lambda item: client.describe_cache_clusters(**item),
      boto_params={
        "ShowCacheNodeInfo": True,
        "ShowCacheClustersNotInReplicationGroups": True
      },
      boto_nextkey = "Marker",
      boto_key="CacheClusters"
    )

  async def _process_account_data_region(self, account, region, resource_groups, loop, *args, **kwargs):
    client = self.get_cloud_client().get_boto_client(client= 'memorydb',  account=account, region=region)
    client_elasticache = self.get_cloud_client().get_boto_client(client= 'elasticache',  account=account, region=region)
    resourcegroupstaggingapi_client = self.get_cloud_client().get_boto_client(client= 'resourcegroupstaggingapi',  account=account, region=region)

    tasks = {
      "clusters": loop.create_task(self.__get_clusters(client= client)),
      "tags": loop.create_task(self.__get_memorydb_tags(client= resourcegroupstaggingapi_client,  account=account, region=region)),
      "elasticache": loop.create_task(self.__get_elasticache(client= client_elasticache)),
    }
    
    if len(tasks) > 0:
      await asyncio.wait(tasks.values())

    elastic_cache_data_tags_tags = {
        cache[self.data_id_name]: loop.create_task(self.__get_elasticache_tags(client= client_elasticache, cache= cache, account= account, region= region)) for cache in tasks["elasticache"].result()
    }
    if len(elastic_cache_data_tags_tags) > 0:
      await asyncio.wait(tasks.values())

    return_data = [
      self.get_common().helper_type().dictionary().merge_dictionary([
        {},
        {
          "extra_tags": tasks["tags"].result().get(item[self.data_id_name].lower()),
          "extra_type": "memorydb"
        }, 
        item
      ]) for item in tasks["clusters"].result()
    ]

    return_data += [
      self.get_common().helper_type().dictionary().merge_dictionary([
        {},
        {
          "extra_tags": elastic_cache_data_tags_tags[item[self.data_id_name]].result() if elastic_cache_data_tags_tags.get(item[self.data_id_name]) is not None else {},
          "extra_type": "elasticache"
        }, 
        item
      ]) for item in tasks["elasticache"].result()
    ]

    return {
      "region": region,
      "resource_groups": resource_groups,
      "data": return_data
    }



from threemystic_cloud_data_client.cloud_providers.base_class.base_data import cloud_data_client_provider_base_data as base
from abc import abstractmethod
import asyncio

class cloud_data_client_aws_client_action_base(base):
  def __init__(self, *args, **kwargs):
    super().__init__(provider= "aws", *args, **kwargs)

  @property
  def skip_global_region(self, *args, **kwargs):
    if hasattr(self, "_skip_global_region"):
      return self._skip_global_region
    
    return True
  
  @skip_global_region.setter
  def skip_global_region(self, value, *args, **kwargs):
    self._skip_global_region = value

  @property
  def override_global_region(self, *args, **kwargs):
    if hasattr(self, "_override_global_region"):
      return self._override_global_region
    
    return None
  
  @override_global_region.setter
  def override_global_region(self, value, *args, **kwargs):
    self._override_global_region = value

  @property
  def auto_region_resourcebytype(self, *args, **kwargs):
    if hasattr(self, "_auto_region_resourcebytype"):
      return self._auto_region_resourcebytype
    
    return []
  
  @auto_region_resourcebytype.setter
  def auto_region_resourcebytype(self, value, *args, **kwargs):
    self._auto_region_resourcebytype = value

  @property  
  def resource_group_filter(self, *args, **kwargs):
    if hasattr(self, "_resource_group_filter"):
      return self._resource_group_filter
    
    return []
  
  @resource_group_filter.setter
  def resource_group_filter(self, value, *args, **kwargs):
    self._resource_group_filter = value
  
  @property
  def arn_lambda(self, *args, **kwargs):
    if hasattr(self, "_arn_lambda"):
      return self._arn_lambda
    
    return lambda item: None
  
  @arn_lambda.setter
  def arn_lambda(self, value, *args, **kwargs):
    self._arn_lambda = value
  
  @property
  def data_id_name(self):
    if hasattr(self, "_data_id_name"):
      return self._data_id_name
    
    return None
  
  @data_id_name.setter
  def data_id_name(self, value):
    self._data_id_name = value

  def get_accounts(self, *args, **kwargs):

    return self.get_cloud_client().get_accounts()
  
  @abstractmethod
  async def _process_account_data_region(self, account, region, resource_groups, loop, *args, **kwargs):
    pass
  
  async def _process_account_region(self, account, region, loop, *args, **kwargs):
    resource_groups = self.__process_account_region_rg(account= account, region= region, loop= loop)
    return await self._process_account_data_region(
      account= account,
      region= region,
      resource_groups= resource_groups if resource_groups is not None else {}, 
      loop= loop,
      **kwargs
    )

  def __process_account_region_rg(self, account, region, loop, *args, **kwargs):
    rg_client = self.get_cloud_client().get_boto_client(
        client= "resource-groups",
        account= account,
        region= region
    )

    resource_groups_by_resource = {}
    
    if self.resource_group_filter is None:
      return resource_groups_by_resource
    
    if len(self.resource_group_filter) < 1:
      return resource_groups_by_resource

    resource_groups = self.get_cloud_client().get_resource_groups(account=account, region=region, rg_client=rg_client)
    
    if len(resource_groups) > 0:
      for filter_item in self.resource_group_filter:
        for resource_id, groups in self.get_cloud_client().get_resource_group_from_resource(account=account, region=region, rg_client=rg_client, resource_groups=resource_groups, filters_resource=[filter_item]).items():
          resource_id = resource_id.lower()
          if resource_id not in resource_groups_by_resource:
            resource_groups_by_resource[resource_id] = groups
            continue

          resource_groups_by_resource[resource_id] += groups
      
      return resource_groups_by_resource

  async def _process_account_data(self, account, loop, *args, **kwargs):

    regions = self.get_cloud_client().get_accounts_regions_costexplorer(
      accounts= [account],
      services= self.auto_region_resourcebytype
    ) if self.auto_region_resourcebytype is not None else {self.get_cloud_client().get_account_id(account= account): []}

    return_data = {
      "account": account,
      "data": [  ]
    }
    if self.get_cloud_client().get_account_id(account= account) not in regions:
      return return_data

    region_tasks = []

    for region_key, region in regions[self.get_cloud_client().get_account_id(account= account)].items():
      if region_key == "global" and self.skip_global_region:
        continue
      
      if self.get_common().helper_type().string().is_null_or_whitespace(string_value= self.override_global_region):
        region_tasks.append(loop.create_task(self._process_account_region(account=account, region=region, loop=loop, **kwargs)))
        continue
      
      if region_key != "global":
        region_tasks.append(loop.create_task(self._process_account_region(account=account, region=region, loop=loop, **kwargs)))
        continue

      if self.get_common().helper_type().string().set_case(string_value= self.override_global_region, case= "lower") in regions[self.get_cloud_client().get_account_id(account= account)]:
        continue

      region_tasks.append(loop.create_task(self._process_account_region(account=account, region= self.override_global_region, loop=loop, **kwargs)))


    if len(region_tasks)>0:
      await asyncio.wait(region_tasks)
    
    for region_task in region_tasks:
      if region_task.result() is None:
        continue

      for item in region_task.result().get("data"):
        resource_arn = self.arn_lambda(
          {
            "region": region_task.result().get("region"),
            "account_id": self.get_cloud_client().get_account_id(account= account),
            "resource_id": item.get(self.data_id_name),
            "raw_item": item
          }
        )
        return_data["data"].append(
          self.get_common().helper_type().dictionary().merge_dictionary([
            {},
            await self.get_base_return_data(
              account= account,
              resource_id= resource_arn,
              resource= item,
              region= region_task.result().get("region"),
              resource_groups= region_task.result().get("resource_groups").get(resource_arn) if region_task.result().get("resource_groups").get(resource_arn) is not None else [],
            ),
            {
              "extra_id_only": item.get(self.data_id_name)
            }
          ])
        )
    return return_data
